import asyncio
import json
import os
import logging
from unittest.mock import patch

from kumiho_memory.summarization import MemorySummarizer


class StubAdapter:
    """Minimal LLMAdapter that returns a canned response."""

    def __init__(self, response: str) -> None:
        self._response = response

    async def chat(self, *, messages, model, system="", max_tokens=1024, json_mode=False):
        return self._response


class StubResponsesClient:
    def __init__(self, output_text: str) -> None:
        self.output_text = output_text
        self.responses = self
        self.chat = type("ChatNamespace", (), {
            "completions": type("CompletionNamespace", (), {
                "create": self._unexpected_chat_completion_call,
            })()
        })()
        self.last_kwargs = None

    async def create(self, **kwargs):
        self.last_kwargs = kwargs
        return type("StubResponse", (), {"output_text": self.output_text, "output": []})()

    async def _unexpected_chat_completion_call(self, **kwargs):
        raise AssertionError(f"chat.completions.create should not be called: {kwargs}")


class StubChatCompletionsClient:
    def __init__(
        self,
        *,
        fail_on_param: str | None = None,
        empty_content: bool = False,
        parsed_payload=None,
        refusal=None,
    ) -> None:
        self.responses = type("ResponsesNamespace", (), {
            "create": self._unexpected_responses_call,
        })()
        self.chat = type("ChatNamespace", (), {
            "completions": type("CompletionNamespace", (), {
                "create": self._create_chat_completion,
            })()
        })()
        self.fail_on_param = fail_on_param
        self.empty_content = empty_content
        self.parsed_payload = parsed_payload
        self.refusal = refusal
        self.calls = []

    async def _create_chat_completion(self, **kwargs):
        self.calls.append(kwargs)
        if self.fail_on_param and self.fail_on_param in kwargs:
            raise ValueError(
                f"Unsupported parameter: '{self.fail_on_param}' is not supported with this model. "
                "Use 'max_completion_tokens' instead."
            )
        response_format = kwargs.get("response_format", {})
        schema = response_format.get("json_schema", {}).get("schema", {})
        schema_type = schema.get("type")
        properties = schema.get("properties", {})
        if "queries" in properties:
            content = '{"queries":["one","two"]}'
        elif "implications" in properties:
            content = '{"implications":["one","two"]}'
        elif schema_type == "array":
            content = '["one","two"]'
        elif schema_type == "object":
            content = '{"summary":"Structured JSON"}'
        else:
            content = kwargs.get("messages", [{}])[-1].get("content", "")
        if self.empty_content:
            content = None
        return type("StubResponse", (), {
            "choices": [type("StubChoice", (), {
                "message": type("StubMessage", (), {
                    "content": (
                        f"ok:{content}"
                        if isinstance(content, str) and not content.startswith("{") and not content.startswith("[")
                        else content
                    ),
                    "parsed": self.parsed_payload,
                    "refusal": self.refusal,
                    "tool_calls": [],
                })(),
                "finish_reason": "stop",
            })()]
        })()

    async def _unexpected_responses_call(self, **kwargs):
        raise AssertionError(f"responses.create should not be called: {kwargs}")


def test_summarize_conversation_with_stubbed_llm():
    canned = json.dumps({
        "type": "summary",
        "title": "Async preference",
        "summary": "User prefers async communication.",
        "knowledge": {"facts": [], "decisions": [], "actions": [], "open_questions": []},
        "classification": {"topics": ["communication"], "entities": []},
    })
    summarizer = MemorySummarizer(adapter=StubAdapter(canned), model="stub")

    messages = [
        {"role": "user", "content": "I prefer async communication."},
        {"role": "assistant", "content": "Understood."},
    ]

    result = asyncio.run(summarizer.summarize_conversation(messages))
    assert result["type"] == "summary"
    assert "communication" in result["summary"].lower()


def test_extract_topics_with_stubbed_llm():
    summarizer = MemorySummarizer(adapter=StubAdapter("memory, agents, redis"), model="stub")

    topics = asyncio.run(summarizer.extract_topics("Memory systems for agents."))
    assert topics == ["memory", "agents", "redis"]


def test_generate_implications_accepts_wrapped_array_object():
    summarizer = MemorySummarizer(
        adapter=StubAdapter('{"implications":["future travel planning","budget pressure"]}'),
        model="stub",
        light_model="stub-light",
    )

    implications = asyncio.run(
        summarizer.generate_implications(
            [{"role": "user", "content": "I need to save more for trips."}],
        )
    )

    assert implications == ["future travel planning", "budget pressure"]


def test_generate_implications_accepts_fenced_wrapped_object():
    summarizer = MemorySummarizer(
        adapter=StubAdapter(
            '```json\n{"implications":["future travel planning","budget pressure"]}\n```'
        ),
        model="stub",
        light_model="stub-light",
    )

    implications = asyncio.run(
        summarizer.generate_implications(
            [{"role": "user", "content": "I need to save more for trips."}],
        )
    )

    assert implications == ["future travel planning", "budget pressure"]


def test_summarize_conversation_includes_debug_payload_on_invalid_json():
    summarizer = MemorySummarizer(
        adapter=StubAdapter("not valid json at all"),
        provider="openai",
        model="gpt-5-mini",
    )

    result = asyncio.run(
        summarizer.summarize_conversation(
            [{"role": "user", "content": "Summarize this please."}],
        )
    )

    assert result["error"] == "No valid JSON found in summarizer response"
    assert result["debug"]["provider"] == "openai"
    assert result["debug"]["model"] == "gpt-5-mini"
    assert result["debug"]["raw_response_len"] == len("not valid json at all")
    assert result["debug"]["raw_response_preview"] == "not valid json at all"


def test_summarize_conversation_logs_diagnostics_on_invalid_json(caplog):
    summarizer = MemorySummarizer(
        adapter=StubAdapter("not valid json at all"),
        provider="openai",
        model="gpt-5-mini",
    )

    with caplog.at_level(logging.WARNING, logger="kumiho_memory.summarization"):
        asyncio.run(
            summarizer.summarize_conversation(
                [{"role": "user", "content": "Summarize this please."}],
            )
        )

    assert "summarize_conversation failed: No valid JSON found in summarizer response" in caplog.text
    assert "provider=openai" in caplog.text
    assert "model=gpt-5-mini" in caplog.text
    assert "raw_preview='not valid json at all'" in caplog.text


def test_custom_adapter_protocol():
    """Verify any object with a chat() method works as an adapter."""

    class GeminiStub:
        async def chat(self, *, messages, model, system="", max_tokens=1024, json_mode=False):
            return json.dumps({
                "type": "fact",
                "title": "Gemini works",
                "summary": "Custom adapters are supported.",
                "knowledge": {"facts": [], "decisions": [], "actions": [], "open_questions": []},
                "classification": {"topics": ["testing"], "entities": []},
            })

    summarizer = MemorySummarizer(adapter=GeminiStub(), model="gemini-2.0-flash")
    assert summarizer.provider == "custom"

    messages = [{"role": "user", "content": "Does Gemini work?"}]
    result = asyncio.run(summarizer.summarize_conversation(messages))
    assert result["type"] == "fact"
    assert result["classification"]["topics"] == ["testing"]


def test_openai_compat_adapter_uses_responses_api_for_codex_models():
    from kumiho_memory.summarization import OpenAICompatAdapter

    client = StubResponsesClient('{"summary":"Codex summary"}')
    adapter = OpenAICompatAdapter(client)

    result = asyncio.run(
        adapter.chat(
            messages=[{"role": "user", "content": "Summarize this"}],
            model="gpt-5-codex",
            system="Return JSON",
            json_mode=True,
        )
    )

    assert result == '{"summary":"Codex summary"}'
    assert client.last_kwargs["model"] == "gpt-5-codex"
    assert "System: Return JSON" in client.last_kwargs["input"]
    assert "User: Summarize this" in client.last_kwargs["input"]


def test_openai_compat_adapter_uses_max_completion_tokens_for_gpt5_chat_models():
    from kumiho_memory.summarization import OpenAICompatAdapter, build_summary_schema_mode

    client = StubChatCompletionsClient()
    adapter = OpenAICompatAdapter(client)

    result = asyncio.run(
        adapter.chat(
            messages=[{"role": "user", "content": "Summarize this"}],
            model="gpt-5.4",
            max_tokens=321,
            json_mode=build_summary_schema_mode(),
        )
    )

    assert result == '{"summary":"Structured JSON"}'
    assert client.calls[0]["max_completion_tokens"] == 321
    assert client.calls[0]["reasoning_effort"] == "none"
    assert "max_tokens" not in client.calls[0]
    assert client.calls[0]["response_format"]["type"] == "json_schema"
    schema = client.calls[0]["response_format"]["json_schema"]["schema"]
    assert schema["type"] == "object"
    assert schema["additionalProperties"] is False
    assert "summary" in schema["properties"]


def test_openai_compat_adapter_uses_json_schema_for_array_mode():
    from kumiho_memory.summarization import OpenAICompatAdapter, build_string_array_wrapper_schema

    client = StubChatCompletionsClient()
    adapter = OpenAICompatAdapter(client)

    result = asyncio.run(
        adapter.chat(
            messages=[{"role": "user", "content": "Return items"}],
            model="gpt-5-mini",
            max_tokens=123,
            json_mode=build_string_array_wrapper_schema("kumiho_queries_response", "queries"),
        )
    )

    assert result == '{"queries":["one","two"]}'
    assert client.calls[0]["response_format"]["type"] == "json_schema"
    assert client.calls[0]["reasoning_effort"] == "minimal"
    schema = client.calls[0]["response_format"]["json_schema"]["schema"]
    assert schema["type"] == "object"
    assert schema["additionalProperties"] is False
    assert schema["properties"]["queries"]["type"] == "array"


def test_openai_compat_adapter_uses_json_object_for_non_native_schema_mode():
    from kumiho_memory.summarization import OpenAICompatAdapter, build_summary_schema_mode

    client = StubChatCompletionsClient()
    adapter = OpenAICompatAdapter(
        client,
        base_url="https://generativelanguage.googleapis.com/v1beta/openai/",
    )

    asyncio.run(
        adapter.chat(
            messages=[{"role": "user", "content": "Summarize this"}],
            model="gemini-2.5-flash",
            max_tokens=321,
            json_mode=build_summary_schema_mode(),
        )
    )

    assert client.calls[0]["response_format"] == {"type": "json_object"}


def test_summarize_conversation_accepts_fenced_json_object():
    canned = (
        "```json\n"
        "{"
        '"type":"summary",'
        '"title":"Async preference",'
        '"summary":"User prefers async communication.",'
        '"events":[],'
        '"knowledge":{"facts":[],"decisions":[],"actions":[],"open_questions":[]},'
        '"classification":{"topics":["communication"],"entities":[]}'
        "}\n"
        "```"
    )
    summarizer = MemorySummarizer(adapter=StubAdapter(canned), model="stub")

    result = asyncio.run(
        summarizer.summarize_conversation(
            [{"role": "user", "content": "I prefer async communication."}],
        )
    )

    assert result["type"] == "summary"
    assert result["summary"] == "User prefers async communication."


def test_openai_compat_adapter_retries_with_max_completion_tokens_when_max_tokens_is_rejected():
    from kumiho_memory.summarization import OpenAICompatAdapter

    client = StubChatCompletionsClient(fail_on_param="max_tokens")
    adapter = OpenAICompatAdapter(client)

    result = asyncio.run(
        adapter.chat(
            messages=[{"role": "user", "content": "Summarize this"}],
            model="gpt-4o",
            max_tokens=222,
        )
    )

    assert result == "ok:Summarize this"
    assert len(client.calls) == 2
    assert client.calls[0]["max_tokens"] == 222
    assert client.calls[1]["max_completion_tokens"] == 222


def test_openai_compat_adapter_uses_parsed_payload_when_content_is_empty():
    from kumiho_memory.summarization import OpenAICompatAdapter, build_summary_schema_mode

    client = StubChatCompletionsClient(
        empty_content=True,
        parsed_payload={"summary": "Structured JSON via parsed"},
    )
    adapter = OpenAICompatAdapter(client)

    result = asyncio.run(
        adapter.chat(
            messages=[{"role": "user", "content": "Summarize this"}],
            model="gpt-5-mini",
            max_tokens=111,
            json_mode=build_summary_schema_mode(),
        )
    )

    assert result == '{"summary": "Structured JSON via parsed"}'


def test_memory_summarizer_adapter_uses_late_kumiho_llm_env():
    from kumiho_memory.summarization import OpenAICompatAdapter

    with patch.dict(os.environ, {}, clear=True):
        summarizer = MemorySummarizer(provider="openai", model="gpt-5-codex")
        stub_adapter = StubAdapter("{}")

        with patch.object(OpenAICompatAdapter, "create", return_value=stub_adapter) as create:
            os.environ["KUMIHO_LLM_API_KEY"] = "late-key"
            os.environ["KUMIHO_LLM_BASE_URL"] = "http://localhost:11434/v1"
            assert summarizer.adapter is stub_adapter

    create.assert_called_once_with(
        api_key="late-key",
        base_url="http://localhost:11434/v1",
    )


def test_normalize_summary_marks_missing_summary_as_error():
    result = MemorySummarizer._normalize_summary(
        {
            "type": "summary",
            "title": "Missing body",
            "knowledge": {},
            "classification": {},
        },
        fallback_messages=[{"role": "user", "content": "We moved the launch to April 15."}],
    )

    assert result["summary"] == "We moved the launch to April 15."
    assert result["error"] == "Summarizer response did not include a non-empty summary"


def test_fallback_summary_ignores_injected_kumiho_blocks():
    messages = [
        {"role": "user", "content": "I replaced the phone battery last week."},
        {
            "role": "assistant",
            "content": (
                "<kumiho_memory>\n"
                "1. Previous note\n"
                "</kumiho_memory>\n"
            ),
        },
    ]

    assert (
        MemorySummarizer._fallback_summary_text(messages)
        == "I replaced the phone battery last week."
    )


def test_fallback_summary_skips_non_user_assistant_messages():
    messages = [
        {"role": "system", "content": "<kumiho_memory>Injected context</kumiho_memory>"},
        {"role": "assistant", "content": "We decided to postpone the migration until April."},
    ]

    assert (
        MemorySummarizer._fallback_summary_text(messages)
        == "We decided to postpone the migration until April."
    )



"""Tests for kumiho_memory.dream_state — DreamState consolidation processor."""

import asyncio
import json
import sys
import tempfile
import types
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional
from unittest.mock import MagicMock

from kumiho_memory.dream_state import (
    DreamState,
    DreamStateStats,
    MemoryAssessment,
    _parse_assessments,
)


# ---------------------------------------------------------------------------
# Stubs — lightweight fakes for the kumiho SDK objects
# ---------------------------------------------------------------------------


@dataclass
class FakeKref:
    uri: str


@dataclass
class FakeRevision:
    kref: FakeKref
    item_kref: FakeKref
    metadata: Dict[str, str] = field(default_factory=dict)
    deprecated: bool = False
    created_at: str = "2026-03-08T12:00:00+00:00"


class FakeItem:
    def __init__(self, kref_uri: str, *, has_get_members: bool = False):
        self.kref = FakeKref(kref_uri)
        self._members: list = []
        self._revisions: list = []
        self._latest_rev: Optional[FakeRevision] = None
        self._has_get_members = has_get_members

    def create_revision(self, metadata: dict):
        rev = FakeRevision(
            kref=FakeKref(f"{self.kref.uri}?r={len(self._revisions) + 1}"),
            item_kref=self.kref,
            metadata=metadata,
        )
        self._revisions.append(rev)
        return _RevisionHandle(rev)

    def get_revision_by_tag(self, tag: str):
        if tag == "latest" and self._latest_rev is not None:
            return self._latest_rev
        if self._revisions:
            return self._revisions[-1]
        return None

    def get_revisions(self):
        return list(self._revisions)

    def get_members(self):
        return self._members


class _RevisionHandle:
    def __init__(self, rev: FakeRevision):
        self.kref = rev.kref
        self._artifacts: list = []

    def create_artifact(self, name: str, path: str):
        self._artifacts.append((name, path))


class FakeSpace:
    def __init__(self, path: str):
        self.path = path


class FakePagedList(list):
    def __init__(self, items, next_cursor=None):
        super().__init__(items)
        self.next_cursor = next_cursor


class FakeClient:
    """Tracks all mutation calls for assertion."""

    def __init__(self):
        self.deprecated: List[str] = []
        self.tags: List[tuple] = []
        self.metadata_updates: List[tuple] = []
        self.edges: List[tuple] = []
        self._published_krefs: set = set()
        self._items_by_space: Dict[str, List[FakeItem]] = {}

    def set_deprecated(self, kref, value):
        self.deprecated.append((kref.uri if hasattr(kref, "uri") else str(kref), value))

    def tag_revision(self, kref, tag):
        self.tags.append((kref.uri if hasattr(kref, "uri") else str(kref), tag))

    def update_revision_metadata(self, kref, updates):
        self.metadata_updates.append(
            (kref.uri if hasattr(kref, "uri") else str(kref), updates)
        )

    def create_edge(self, source, target, edge_type):
        src = source.kref.uri if hasattr(source, "kref") else str(source)
        tgt = target.kref.uri if hasattr(target, "kref") else str(target)
        self.edges.append((src, tgt, edge_type))

    def has_tag(self, kref, tag):
        uri = kref.uri if hasattr(kref, "uri") else str(kref)
        return uri in self._published_krefs and tag == "published"

    def get_items(self, parent_path="", kind_filter="", include_deprecated=False):
        return self._items_by_space.get(parent_path, [])


def _build_fake_sdk(
    *,
    revisions: Optional[List[FakeRevision]] = None,
    items: Optional[Dict[str, FakeItem]] = None,
    attributes: Optional[Dict[str, Dict[str, str]]] = None,
    client: Optional[FakeClient] = None,
    spaces: Optional[List[FakeSpace]] = None,
    items_by_space: Optional[Dict[str, List[FakeItem]]] = None,
):
    """Create a fake ``kumiho`` module that mimics the real SDK."""
    revisions = revisions or []
    items = items or {}
    attributes = attributes if attributes is not None else {}
    client = client or FakeClient()
    spaces = spaces or [FakeSpace("/CognitiveMemory/personal")]

    if items_by_space:
        client._items_by_space = items_by_space

    sdk = types.ModuleType("kumiho")

    def get_item(kref_uri):
        return items.get(kref_uri)

    def get_project(name):
        proj = types.SimpleNamespace()
        proj.name = name
        proj.get_space = lambda n: types.SimpleNamespace(
            create_item=lambda name, kind: _ensure_item(items, f"kref://{name}/{n}.{kind}")
        )
        proj.create_space = lambda n: types.SimpleNamespace(
            create_item=lambda name, kind: _ensure_item(items, f"kref://{name}/{n}.{kind}")
        )
        proj.get_spaces = lambda recursive=False: list(spaces)
        return proj

    def get_attribute(kref, key):
        return attributes.get(kref, {}).get(key)

    def set_attribute(kref, key, value):
        attributes.setdefault(kref, {})[key] = value

    def get_revision(kref_str):
        for rev in revisions:
            if rev.kref.uri == kref_str:
                return rev
        # Return a simple stub so create_edge doesn't fail
        return types.SimpleNamespace(kref=FakeKref(kref_str))

    def get_client_fn():
        return client

    sdk.get_item = get_item
    sdk.get_project = get_project
    sdk.get_attribute = get_attribute
    sdk.set_attribute = set_attribute
    sdk.get_revision = get_revision
    sdk.get_client = get_client_fn
    sdk.Kref = FakeKref

    return sdk, client, attributes


def _ensure_item(items: dict, kref_uri: str) -> FakeItem:
    if kref_uri not in items:
        items[kref_uri] = FakeItem(kref_uri)
    return items[kref_uri]


class StubSummarizer:
    """A MemorySummarizer stand-in with a controllable adapter."""

    def __init__(self, response: str = "[]"):
        self.adapter = StubAdapter(response)
        self.model = "stub-model"


class StubAdapter:
    """LLMAdapter stand-in that returns pre-configured JSON."""

    def __init__(self, response: str = "[]"):
        self._response = response

    async def chat(self, *, messages, model, system, max_tokens, json_mode=False):
        return self._response


# ---------------------------------------------------------------------------
# Helper to inject the fake SDK
# ---------------------------------------------------------------------------


def _make_dream_state(sdk_module, summarizer=None, **kwargs):
    """Build a DreamState and monkey-patch its kumiho import."""
    ds = DreamState(
        summarizer=summarizer or StubSummarizer(),
        **kwargs,
    )
    # Patch `import kumiho` inside run() by injecting into sys.modules
    sys.modules["kumiho"] = sdk_module
    return ds


def _cleanup_sdk():
    sys.modules.pop("kumiho", None)


def _make_item_with_revision(
    item_kref: str,
    rev_kref: str,
    metadata: Dict[str, str],
    created_at: str = "2026-03-08T12:00:00+00:00",
    deprecated: bool = False,
) -> tuple:
    """Create a FakeItem with a latest revision pre-attached."""
    item = FakeItem(item_kref)
    rev = FakeRevision(
        kref=FakeKref(rev_kref),
        item_kref=FakeKref(item_kref),
        metadata=metadata,
        deprecated=deprecated,
        created_at=created_at,
    )
    item._latest_rev = rev
    return item, rev


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------


def test_run_no_revisions():
    """No revisions since last run → report with zeros."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}
    # No items in any space
    sdk, client, attrs = _build_fake_sdk(
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={},
    )

    async def run():
        ds = _make_dream_state(sdk)
        ds._cursor_item_kref = cursor_item.kref.uri
        try:
            report = await ds.run()
            assert report["success"] is True
            assert report["events_processed"] == 0
            assert report["revisions_assessed"] == 0
            assert report["deprecated"] == 0
        finally:
            _cleanup_sdk()

    asyncio.run(run())


def test_collect_revisions_uses_paginated_space_and_item_listing():
    """Dream State should avoid one-shot recursive listing in large projects."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/new.conversation",
        "kref://CognitiveMemory/personal/new.conversation?r=1",
        {"title": "New memory", "summary": "After pagination"},
        created_at="2026-03-08T14:00:00+00:00",
    )

    fake_client = FakeClient()
    item_calls = []

    def paged_get_items(
        parent_path="",
        kind_filter="",
        page_size=None,
        cursor=None,
        include_deprecated=False,
    ):
        item_calls.append((parent_path, page_size, cursor))
        if parent_path == "/CognitiveMemory":
            return FakePagedList([], next_cursor=None)
        if cursor is None:
            return FakePagedList([mem_item], next_cursor="page-2")
        return FakePagedList([], next_cursor=None)

    fake_client.get_items = paged_get_items

    class FakeProject:
        def __init__(self):
            self.space_calls = []

        def get_spaces(
            self,
            parent_path=None,
            recursive=False,
            page_size=None,
            cursor=None,
        ):
            self.space_calls.append((parent_path, recursive, page_size, cursor))
            if parent_path == "/CognitiveMemory":
                return FakePagedList([FakeSpace("/CognitiveMemory/personal")])
            return FakePagedList([])

    project = FakeProject()

    sdk = types.ModuleType("kumiho")
    sdk.get_project = lambda name: project
    sdk.get_client = lambda: fake_client
    sdk.get_item = lambda kref_uri: items.get(kref_uri)
    sdk.get_attribute = lambda kref, key: None
    sdk.set_attribute = lambda kref, key, value: None
    sdk.get_revision = lambda kref_str: rev
    sdk.Kref = FakeKref

    ds = _make_dream_state(sdk)
    ds._cursor_item_kref = cursor_item.kref.uri
    try:
        revisions = ds._collect_revisions(sdk, None)
        assert revisions == [rev]
        assert project.space_calls[0] == ("/CognitiveMemory", False, 100, None)
        assert ("/CognitiveMemory/personal", 100, None) in item_calls
        assert ("/CognitiveMemory/personal", 100, "page-2") in item_calls
    finally:
        _cleanup_sdk()


def test_run_processes_revisions():
    """Revisions found → assess → apply full pipeline."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/mem1.conversation",
        "kref://CognitiveMemory/personal/mem1.conversation?r=1",
        {"title": "User preference", "summary": "User likes dark mode"},
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.8,
        "should_deprecate": False,
        "deprecation_reason": "",
        "suggested_tags": ["preference", "ui"],
        "metadata_updates": {"topic": "dark-mode"},
        "related_indices": [],
        "relationship_type": "",
    }])

    sdk, client, attrs = _build_fake_sdk(
        revisions=[rev],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": [mem_item]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                assert report["events_processed"] == 1
                assert report["revisions_assessed"] == 1
                assert report["deprecated"] == 0
                assert report["tags_added"] == 2
                assert report["metadata_updated"] == 1
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_load_last_run_at_first_run():
    """Returns None when no timestamp exists."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}
    sdk, _, attrs = _build_fake_sdk(items=items)

    with tempfile.TemporaryDirectory() as tmpdir:
        ds = _make_dream_state(sdk, artifact_root=tmpdir)
        try:
            result = ds._load_last_run_at(sdk, cursor_item.kref.uri)
            assert result is None
        finally:
            _cleanup_sdk()


def test_save_and_load_last_run_at():
    """Round-trip timestamp persistence."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}
    sdk, _, attrs = _build_fake_sdk(items=items)

    with tempfile.TemporaryDirectory() as tmpdir:
        ds = _make_dream_state(sdk, artifact_root=tmpdir)
        try:
            assert ds._load_last_run_at(sdk, cursor_item.kref.uri) is None

            ds._save_last_run_at(
                sdk,
                cursor_item.kref.uri,
                "2026-03-08T10:00:00+00:00",
            )
            loaded = ds._load_last_run_at(sdk, cursor_item.kref.uri)
            assert loaded == "2026-03-08T10:00:00+00:00"
        finally:
            _cleanup_sdk()


def test_revision_time_filter():
    """Only revisions after last_run_at are collected."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    # Old revision — should be skipped
    old_item, old_rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/old.conversation",
        "kref://CognitiveMemory/personal/old.conversation?r=1",
        {"title": "Old memory", "summary": "Before cutoff"},
        created_at="2026-03-07T08:00:00+00:00",
    )

    # New revision — should be collected
    new_item, new_rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/new.conversation",
        "kref://CognitiveMemory/personal/new.conversation?r=1",
        {"title": "New memory", "summary": "After cutoff"},
        created_at="2026-03-08T14:00:00+00:00",
    )

    sdk, client, attrs = _build_fake_sdk(
        revisions=[old_rev, new_rev],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": [old_item, new_item]},
        # Set last_run_at so old items are filtered out
        attributes={
            cursor_item.kref.uri: {
                "last_run_at": "2026-03-08T10:00:00+00:00",
            }
        },
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.8,
        "should_deprecate": False,
        "deprecation_reason": "",
        "suggested_tags": ["new"],
        "metadata_updates": {},
        "related_indices": [],
        "relationship_type": "",
    }])

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                # Only the new revision should be processed
                assert report["events_processed"] == 1
                assert report["revisions_assessed"] == 1
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_assess_deprecation():
    """LLM recommends deprecation → set_deprecated called."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/old.conversation",
        "kref://CognitiveMemory/personal/old.conversation?r=1",
        {"title": "Old info", "summary": "Outdated data"},
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.1,
        "should_deprecate": True,
        "deprecation_reason": "outdated information",
        "suggested_tags": [],
        "metadata_updates": {},
        "related_indices": [],
        "relationship_type": "",
    }])

    sdk, fake_client, attrs = _build_fake_sdk(
        revisions=[rev],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": [mem_item]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                assert report["deprecated"] == 1
                assert len(fake_client.deprecated) == 1
                assert fake_client.deprecated[0][1] is True
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_assess_tag_updates():
    """LLM recommends tags → tag_revision called."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/work/task.conversation",
        "kref://CognitiveMemory/work/task.conversation?r=1",
        {"title": "CI pipeline", "summary": "Setup GitHub Actions"},
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.9,
        "should_deprecate": False,
        "deprecation_reason": "",
        "suggested_tags": ["ci-cd", "github-actions", "devops"],
        "metadata_updates": {},
        "related_indices": [],
        "relationship_type": "",
    }])

    sdk, fake_client, _ = _build_fake_sdk(
        revisions=[rev],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/work")],
        items_by_space={"/CognitiveMemory/work": [mem_item]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["tags_added"] == 3
                tag_names = [t[1] for t in fake_client.tags]
                assert "ci-cd" in tag_names
                assert "github-actions" in tag_names
                assert "devops" in tag_names
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_assess_relationships():
    """LLM recommends edges → create_edge called."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    kref_a = "kref://CognitiveMemory/work/a.conversation"
    kref_b = "kref://CognitiveMemory/work/b.conversation"

    item_a, rev_a = _make_item_with_revision(
        kref_a, f"{kref_a}?r=1",
        {"title": "Deploy v1", "summary": "First deploy"},
    )
    item_b, rev_b = _make_item_with_revision(
        kref_b, f"{kref_b}?r=1",
        {"title": "Deploy v2", "summary": "Second deploy"},
    )

    llm_response = json.dumps([
        {
            "index": 0,
            "relevance_score": 0.8,
            "should_deprecate": False,
            "deprecation_reason": "",
            "suggested_tags": [],
            "metadata_updates": {},
            "related_indices": [1],
            "relationship_type": "DERIVED_FROM",
        },
        {
            "index": 1,
            "relevance_score": 0.9,
            "should_deprecate": False,
            "deprecation_reason": "",
            "suggested_tags": [],
            "metadata_updates": {},
            "related_indices": [],
            "relationship_type": "",
        },
    ])

    sdk, fake_client, _ = _build_fake_sdk(
        revisions=[rev_a, rev_b],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/work")],
        items_by_space={"/CognitiveMemory/work": [item_a, item_b]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["edges_created"] == 1
                assert len(fake_client.edges) == 1
                _, _, edge_type = fake_client.edges[0]
                assert edge_type == "DERIVED_FROM"
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_dry_run_no_mutations():
    """dry_run=True → no SDK mutation calls."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/dry.conversation",
        "kref://CognitiveMemory/personal/dry.conversation?r=1",
        {"title": "Dry run test", "summary": "Should not mutate"},
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.1,
        "should_deprecate": True,
        "deprecation_reason": "test deprecation",
        "suggested_tags": ["test-tag"],
        "metadata_updates": {"key": "val"},
        "related_indices": [],
        "relationship_type": "",
    }])

    sdk, fake_client, attrs = _build_fake_sdk(
        revisions=[rev],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": [mem_item]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                dry_run=True,
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                assert report["revisions_assessed"] == 1
                # No mutations should have occurred
                assert len(fake_client.deprecated) == 0
                assert len(fake_client.tags) == 0
                assert len(fake_client.metadata_updates) == 0
                assert len(fake_client.edges) == 0
                # Stats should still show 0 since apply was skipped
                assert report["deprecated"] == 0
                assert report["tags_added"] == 0
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_published_never_deprecated():
    """Published items skipped even if LLM says deprecate."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    kref_str = "kref://CognitiveMemory/personal/pub.conversation?r=1"
    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/pub.conversation",
        kref_str,
        {"title": "Published doc", "summary": "Important published data"},
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.1,
        "should_deprecate": True,
        "deprecation_reason": "LLM says deprecate",
        "suggested_tags": [],
        "metadata_updates": {},
        "related_indices": [],
        "relationship_type": "",
    }])

    fake_client = FakeClient()
    # Mark the kref as having the "published" tag
    fake_client._published_krefs.add(kref_str)

    sdk, _, attrs = _build_fake_sdk(
        revisions=[rev],
        items=items,
        client=fake_client,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": [mem_item]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                # Should NOT have deprecated the published item
                assert report["deprecated"] == 0
                assert len(fake_client.deprecated) == 0
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_max_deprecation_guard():
    """Max 50% deprecation circuit breaker."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    # Create 4 items, LLM says deprecate ALL of them
    mem_items = []
    revisions = []
    llm_items = []
    for i in range(4):
        kref_base = f"kref://CognitiveMemory/personal/item{i}.conversation"
        kref_rev = f"{kref_base}?r=1"
        item, rev = _make_item_with_revision(
            kref_base, kref_rev,
            {"title": f"Item {i}", "summary": f"Content {i}"},
        )
        mem_items.append(item)
        revisions.append(rev)
        llm_items.append({
            "index": i,
            "relevance_score": 0.1,
            "should_deprecate": True,
            "deprecation_reason": "not useful",
            "suggested_tags": [],
            "metadata_updates": {},
            "related_indices": [],
            "relationship_type": "",
        })

    llm_response = json.dumps(llm_items)
    sdk, fake_client, _ = _build_fake_sdk(
        revisions=revisions,
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": mem_items},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                # 4 items assessed, max 50% = 2 can be deprecated
                assert report["deprecated"] == 2
                assert len(fake_client.deprecated) == 2
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_report_generation():
    """Report revision + artifact created."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/work/rep.conversation",
        "kref://CognitiveMemory/work/rep.conversation?r=1",
        {"title": "Report test", "summary": "Testing report gen"},
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.7,
        "should_deprecate": False,
        "deprecation_reason": "",
        "suggested_tags": ["testing"],
        "metadata_updates": {},
        "related_indices": [],
        "relationship_type": "",
    }])

    sdk, _, _ = _build_fake_sdk(
        revisions=[rev],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/work")],
        items_by_space={"/CognitiveMemory/work": [mem_item]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                # A report_kref should be returned
                assert "report_kref" in report
                assert report["report_kref"] is not None
                # A revision should have been created on the cursor item
                assert len(cursor_item._revisions) == 1
                rev_meta = cursor_item._revisions[0].metadata
                assert rev_meta["type"] == "dream_state_report"
                assert int(rev_meta["tags_added"]) == 1
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_stacked_revision_detected():
    """A new revision on an existing item (stacked) is picked up."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    # Item already existed with r=1, now has a new stacked revision r=2
    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/existing.conversation",
        "kref://CognitiveMemory/personal/existing.conversation?r=2",
        {"title": "Updated preference", "summary": "User changed to light mode"},
        created_at="2026-03-08T15:00:00+00:00",
    )

    llm_response = json.dumps([{
        "index": 0,
        "relevance_score": 0.9,
        "should_deprecate": False,
        "deprecation_reason": "",
        "suggested_tags": ["preference-updated"],
        "metadata_updates": {},
        "related_indices": [],
        "relationship_type": "",
    }])

    sdk, client, attrs = _build_fake_sdk(
        revisions=[rev],
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": [mem_item]},
        # Last run was before this revision
        attributes={
            cursor_item.kref.uri: {
                "last_run_at": "2026-03-08T10:00:00+00:00",
            }
        },
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(
                sdk,
                summarizer=StubSummarizer(llm_response),
                artifact_root=tmpdir,
            )
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                assert report["events_processed"] == 1
                assert report["revisions_assessed"] == 1
                assert report["tags_added"] == 1
            finally:
                _cleanup_sdk()

        asyncio.run(run())


def test_deprecated_revisions_skipped():
    """Already-deprecated revisions should not be collected."""
    cursor_item = FakeItem("kref://CognitiveMemory/_dream_state.conversation")
    items = {cursor_item.kref.uri: cursor_item}

    mem_item, rev = _make_item_with_revision(
        "kref://CognitiveMemory/personal/dep.conversation",
        "kref://CognitiveMemory/personal/dep.conversation?r=1",
        {"title": "Deprecated", "summary": "Already deprecated"},
        deprecated=True,
    )

    sdk, client, attrs = _build_fake_sdk(
        items=items,
        spaces=[FakeSpace("/CognitiveMemory/personal")],
        items_by_space={"/CognitiveMemory/personal": [mem_item]},
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        async def run():
            ds = _make_dream_state(sdk, artifact_root=tmpdir)
            ds._cursor_item_kref = cursor_item.kref.uri
            try:
                report = await ds.run()
                assert report["success"] is True
                assert report["events_processed"] == 0
            finally:
                _cleanup_sdk()

        asyncio.run(run())


# ---------------------------------------------------------------------------
# Helper function tests
# ---------------------------------------------------------------------------


def test_parse_assessments_valid_json():
    """Direct JSON array should parse correctly."""
    raw = json.dumps([{"index": 0, "relevance_score": 0.5}])
    result = _parse_assessments(raw)
    assert len(result) == 1
    assert result[0]["index"] == 0


def test_parse_assessments_markdown_fenced():
    """JSON inside markdown code fences should be extracted."""
    raw = '```json\n[{"index": 0, "relevance_score": 0.9}]\n```'
    result = _parse_assessments(raw)
    assert len(result) == 1


def test_parse_assessments_wrapped_object():
    """Strict-schema OpenAI responses may wrap the array in an object."""
    raw = json.dumps({
        "assessments": [{"index": 0, "relevance_score": 0.9}],
    })
    result = _parse_assessments(raw)
    assert len(result) == 1
    assert result[0]["index"] == 0


def test_parse_assessments_invalid_returns_empty():
    """Unparseable text should return an empty list."""
    result = _parse_assessments("This is not JSON at all.")
    assert result == []



"""LLM-based summarization for memory consolidation."""

from __future__ import annotations

import json
import logging
import os
import re
from typing import Any, Dict, List, Optional, Protocol, Union, runtime_checkable

logger = logging.getLogger(__name__)

JsonMode = Union[bool, str, Dict[str, Any]]


def _strict_object_schema(
    properties: Dict[str, Any],
    *,
    required: Optional[List[str]] = None,
) -> Dict[str, Any]:
    return {
        "type": "object",
        "additionalProperties": False,
        "properties": properties,
        "required": required or list(properties.keys()),
    }


def _json_schema_mode(name: str, schema: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "name": name,
        "schema": schema,
    }


def build_string_array_wrapper_schema(name: str, field_name: str) -> Dict[str, Any]:
    return _json_schema_mode(
        name,
        _strict_object_schema({
            field_name: {
                "type": "array",
                "items": {"type": "string"},
            },
        }),
    )


def build_summary_schema_mode() -> Dict[str, Any]:
    return _json_schema_mode(
        "kumiho_summary_response",
        _strict_object_schema({
            "type": {"type": "string"},
            "title": {"type": "string"},
            "summary": {"type": "string"},
            "events": {
                "type": "array",
                "items": _strict_object_schema({
                    "event": {"type": "string"},
                    "when": {"type": "string"},
                    "participants": {
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "consequence": {"type": "string"},
                }),
            },
            "knowledge": _strict_object_schema({
                "facts": {
                    "type": "array",
                    "items": _strict_object_schema({
                        "claim": {"type": "string"},
                        "certainty": {"type": "string"},
                    }),
                },
                "decisions": {
                    "type": "array",
                    "items": _strict_object_schema({
                        "decision": {"type": "string"},
                        "reason": {"type": "string"},
                    }),
                },
                "actions": {
                    "type": "array",
                    "items": _strict_object_schema({
                        "task": {"type": "string"},
                        "status": {"type": "string"},
                    }),
                },
                "open_questions": {
                    "type": "array",
                    "items": {"type": "string"},
                },
            }),
            "classification": _strict_object_schema({
                "topics": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "entities": {
                    "type": "array",
                    "items": {"type": "string"},
                },
            }),
        }),
    )


# ---------------------------------------------------------------------------
# LLM Adapter protocol — implement this to add any provider
# ---------------------------------------------------------------------------


@runtime_checkable
class LLMAdapter(Protocol):
    """Interface for LLM providers.

    Implement this protocol to plug in any LLM provider (Gemini, Mistral,
    Ollama, Cohere, etc.).  Built-in adapters are provided for
    OpenAI-compatible and Anthropic APIs.

    Example custom adapter::

        class MyAdapter:
            async def chat(self, *, messages, model, system="",
                           max_tokens=1024, json_mode=False) -> str:
                # call your LLM here
                return response_text
    """

    async def chat(
        self,
        *,
        messages: List[Dict[str, str]],
        model: str,
        system: str = "",
        max_tokens: int = 1024,
        json_mode: JsonMode = False,
    ) -> str:
        """Send a chat request and return the raw response text."""
        ...


# ---------------------------------------------------------------------------
# Built-in adapters
# ---------------------------------------------------------------------------


class OpenAICompatAdapter:
    """Adapter for OpenAI and any OpenAI-compatible API.

    Works with: OpenAI, Azure OpenAI, Google Gemini (via OpenAI compat),
    Ollama, vLLM, LiteLLM, Together, Groq, Mistral, and others.

    Usage::

        # Standard OpenAI
        adapter = OpenAICompatAdapter.create(api_key="sk-...")

        # Gemini via OpenAI-compatible endpoint
        adapter = OpenAICompatAdapter.create(
            api_key="your-gemini-key",
            base_url="https://generativelanguage.googleapis.com/v1beta/openai/",
        )

        # Local Ollama
        adapter = OpenAICompatAdapter.create(
            base_url="http://localhost:11434/v1",
        )
    """

    def __init__(self, client: Any, *, base_url: Optional[str] = None) -> None:
        self._client = client
        self._base_url = (base_url or "").rstrip("/")

    @staticmethod
    def _json_preview(value: Any, *, limit: int = 240) -> str:
        if value is None:
            return ""
        if isinstance(value, str):
            text = value
        else:
            try:
                text = json.dumps(value, default=str)
            except Exception:
                text = repr(value)
        compact = re.sub(r"\s+", " ", text).strip()
        if len(compact) <= limit:
            return compact
        return compact[: limit - 3] + "..."

    @classmethod
    def _coerce_message_content(cls, content: Any) -> str:
        if isinstance(content, str):
            return content
        if isinstance(content, list):
            text_parts: List[str] = []
            for item in content:
                if isinstance(item, str):
                    text_parts.append(item)
                    continue
                text = getattr(item, "text", None)
                if isinstance(text, str) and text:
                    text_parts.append(text)
                    continue
                if isinstance(item, dict):
                    maybe_text = item.get("text")
                    if isinstance(maybe_text, str) and maybe_text:
                        text_parts.append(maybe_text)
            return "\n".join(part for part in text_parts if part).strip()
        return ""

    def _extract_chat_message_text(
        self,
        *,
        response: Any,
        model: str,
        json_mode: JsonMode,
        reasoning_effort: Optional[str] = None,
    ) -> str:
        choices = getattr(response, "choices", None) or []
        if not choices:
            logger.warning(
                "chat.completions returned no choices | model=%s base_url=%s json_mode=%s reasoning_effort=%s",
                model,
                self._base_url or "",
                MemorySummarizer._describe_json_mode(json_mode),
                reasoning_effort or "",
            )
            return ""

        choice = choices[0]
        message = getattr(choice, "message", None)
        content = self._coerce_message_content(getattr(message, "content", None))
        if content:
            return content

        parsed = getattr(message, "parsed", None)
        if parsed is not None:
            try:
                return json.dumps(parsed, default=str)
            except Exception:
                pass

        logger.warning(
            "chat.completions returned empty message content | model=%s base_url=%s json_mode=%s reasoning_effort=%s finish_reason=%s refusal=%r parsed_type=%s parsed_preview=%r content_type=%s content_preview=%r tool_calls=%s",
            model,
            self._base_url or "",
            MemorySummarizer._describe_json_mode(json_mode),
            reasoning_effort or "",
            getattr(choice, "finish_reason", ""),
            getattr(message, "refusal", None),
            type(parsed).__name__ if parsed is not None else "",
            self._json_preview(parsed),
            type(getattr(message, "content", None)).__name__,
            self._json_preview(getattr(message, "content", None)),
            len(getattr(message, "tool_calls", None) or []),
        )
        return ""

    @staticmethod
    def _prefers_max_completion_tokens(model: str) -> bool:
        lowered = model.strip().lower()
        return bool(re.match(r"^(gpt-5|o\d)", lowered))

    @staticmethod
    def _supports_alternate_token_param(exc: Exception, *, failed_param: str) -> bool:
        message = str(exc).lower()
        return (
            "unsupported parameter" in message and
            failed_param in message
        )

    @staticmethod
    def _default_reasoning_effort(model: str, json_mode: JsonMode) -> Optional[str]:
        if not json_mode:
            return None
        lowered = model.strip().lower()
        if lowered.startswith(("gpt-5.1", "gpt-5.2", "gpt-5.4")):
            return "none"
        if lowered.startswith("gpt-5"):
            return "minimal"
        return None

    @staticmethod
    def _json_mode_kind(json_mode: JsonMode) -> Optional[str]:
        if isinstance(json_mode, dict):
            return "schema"
        if json_mode == "array":
            return "array"
        if json_mode:
            return "object"
        return None

    def _is_native_openai(self) -> bool:
        return not self._base_url or "api.openai.com" in self._base_url.lower()

    def _response_format(self, json_mode: JsonMode) -> Optional[Dict[str, Any]]:
        mode_kind = self._json_mode_kind(json_mode)
        if not mode_kind:
            return None

        if self._is_native_openai():
            if isinstance(json_mode, dict):
                return {
                    "type": "json_schema",
                    "json_schema": {
                        "name": json_mode["name"],
                        "strict": True,
                        "schema": json_mode["schema"],
                    },
                }
            return {
                "type": "json_schema",
                "json_schema": {
                    "name": f"kumiho_{mode_kind}_response",
                    "strict": True,
                    "schema": {"type": mode_kind},
                },
            }

        if isinstance(json_mode, dict) or mode_kind == "object":
            return {"type": "json_object"}

        return None

    @classmethod
    def create(
        cls,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
    ) -> "OpenAICompatAdapter":
        try:
            from openai import AsyncOpenAI  # type: ignore
        except ImportError as exc:
            raise ImportError(
                "openai package is required: pip install kumiho-memory[openai]"
            ) from exc

        kwargs: Dict[str, Any] = {}
        if api_key:
            kwargs["api_key"] = api_key
        if base_url:
            kwargs["base_url"] = base_url
        return cls(AsyncOpenAI(**kwargs), base_url=base_url)

    async def chat(
        self,
        *,
        messages: List[Dict[str, str]],
        model: str,
        system: str = "",
        max_tokens: int = 1024,
        json_mode: JsonMode = False,
    ) -> str:
        full_messages: List[Dict[str, str]] = []
        if system:
            full_messages.append({"role": "system", "content": system})
        full_messages.extend(messages)

        if "codex" in model.lower():
            prompt_parts: List[str] = []
            for msg in full_messages:
                role = str(msg.get("role", "user")).strip().capitalize()
                content = str(msg.get("content", "")).strip()
                if content:
                    prompt_parts.append(f"{role}: {content}")

            response = await self._client.responses.create(
                model=model,
                input="\n\n".join(prompt_parts),
                max_output_tokens=max_tokens,
            )

            output_text = getattr(response, "output_text", "") or ""
            if output_text:
                return output_text

            output = getattr(response, "output", None) or []
            text_parts: List[str] = []
            for item in output:
                for content in getattr(item, "content", None) or []:
                    text = getattr(content, "text", None)
                    if text:
                        text_parts.append(str(text))
            joined = "\n".join(part for part in text_parts if part).strip()
            if not joined:
                logger.warning(
                    "responses.create returned no output text | model=%s base_url=%s output_items=%s",
                    model,
                    self._base_url or "",
                    len(output),
                )
            return joined

        kwargs: Dict[str, Any] = {
            "model": model,
            "messages": full_messages,
        }
        reasoning_effort = self._default_reasoning_effort(model, json_mode)
        token_param = (
            "max_completion_tokens"
            if self._prefers_max_completion_tokens(model)
            else "max_tokens"
        )
        kwargs[token_param] = max_tokens
        if reasoning_effort:
            kwargs["reasoning_effort"] = reasoning_effort
        response_format = self._response_format(json_mode)
        if response_format:
            kwargs["response_format"] = response_format

        try:
            response = await self._client.chat.completions.create(**kwargs)
        except Exception as exc:
            alt_param = "max_tokens" if token_param == "max_completion_tokens" else "max_completion_tokens"
            if not self._supports_alternate_token_param(exc, failed_param=token_param):
                raise
            retry_kwargs = dict(kwargs)
            retry_kwargs.pop(token_param, None)
            retry_kwargs[alt_param] = max_tokens
            response = await self._client.chat.completions.create(**retry_kwargs)
        return self._extract_chat_message_text(
            response=response,
            model=model,
            json_mode=json_mode,
            reasoning_effort=reasoning_effort,
        )


class AnthropicAdapter:
    """Adapter for the Anthropic Messages API."""

    def __init__(self, client: Any) -> None:
        self._client = client

    @classmethod
    def create(cls, *, api_key: Optional[str] = None) -> "AnthropicAdapter":
        try:
            from anthropic import AsyncAnthropic  # type: ignore
        except ImportError as exc:
            raise ImportError(
                "anthropic package is required: pip install kumiho-memory[anthropic]"
            ) from exc

        kwargs: Dict[str, Any] = {}
        if api_key:
            kwargs["api_key"] = api_key
        return cls(AsyncAnthropic(**kwargs))

    async def chat(
        self,
        *,
        messages: List[Dict[str, str]],
        model: str,
        system: str = "",
        max_tokens: int = 1024,
        json_mode: JsonMode = False,
    ) -> str:
        kwargs: Dict[str, Any] = {
            "model": model,
            "messages": messages,
            "max_tokens": max_tokens,
        }
        if system:
            kwargs["system"] = system

        response = await self._client.messages.create(**kwargs)
        return response.content[0].text if response.content else ""


# ---------------------------------------------------------------------------
# Embedding adapter protocol — opt-in for embedding-based sibling filtering
# ---------------------------------------------------------------------------


@runtime_checkable
class EmbeddingAdapter(Protocol):
    """Interface for text embedding providers.

    Implement this protocol to plug in any embedding provider.  Used by
    the memory manager for embedding-based sibling relevance filtering.
    """

    def embed(
        self,
        texts: List[str],
        *,
        model: str = "",
    ) -> List[List[float]]:
        """Embed a batch of texts and return a list of float vectors."""
        ...


class OpenAICompatEmbeddingAdapter:
    """Embedding adapter for OpenAI and compatible APIs.

    Usage::

        adapter = OpenAICompatEmbeddingAdapter.create(api_key="sk-...")
        vectors = adapter.embed(["hello world", "test"])
    """

    def __init__(self, client: Any, default_model: str = "text-embedding-3-small") -> None:
        self._client = client
        self._default_model = default_model

    @classmethod
    def create(
        cls,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        model: str = "text-embedding-3-small",
    ) -> "OpenAICompatEmbeddingAdapter":
        try:
            from openai import OpenAI  # type: ignore
        except ImportError as exc:
            raise ImportError(
                "openai package is required: pip install kumiho-memory[openai]"
            ) from exc

        kwargs: Dict[str, Any] = {}
        if api_key:
            kwargs["api_key"] = api_key
        if base_url:
            kwargs["base_url"] = base_url
        return cls(OpenAI(**kwargs), default_model=model)

    def embed(
        self,
        texts: List[str],
        *,
        model: str = "",
    ) -> List[List[float]]:
        resolved_model = model or self._default_model
        resp = self._client.embeddings.create(input=texts, model=resolved_model)
        return [item.embedding for item in resp.data]


# ---------------------------------------------------------------------------
# Provider / model defaults
# ---------------------------------------------------------------------------

PROVIDER_DEFAULTS: Dict[str, Dict[str, str]] = {
    "openai": {
        "model": "gpt-4o",
        "light_model": "gpt-4o-mini",
    },
    "anthropic": {
        "model": "claude-sonnet-4-5-20241022",
        "light_model": "claude-haiku-4-20250414",
    },
}

# Key-prefix heuristics used when the caller only supplies an API key.
_KEY_PREFIX_MAP = {
    "sk-": "openai",
}


def _detect_provider(api_key: str) -> str:
    """Best-effort provider detection from API key prefix."""
    for prefix, provider in _KEY_PREFIX_MAP.items():
        if api_key.startswith(prefix):
            return provider
    # Anthropic keys don't have a stable prefix; default to openai if unsure.
    return "openai"


class MemorySummarizer:
    """Summarize conversations into structured memory.

    Supports any LLM provider through the adapter pattern:

    1. **Built-in providers** (``provider="openai"`` or ``"anthropic"``):
       Auto-configured from env vars or explicit arguments.

    2. **OpenAI-compatible APIs** (Gemini, Ollama, Groq, Together, vLLM…):
       Set ``base_url`` to point at the compatible endpoint.

    3. **Fully custom providers**: Pass any object implementing the
       ``LLMAdapter`` protocol via the ``adapter`` parameter.

    Configuration priority (highest → lowest):

    1. Explicit ``adapter`` — bypasses all other config.
    2. Constructor arguments (``api_key``, ``provider``, ``model``,
       ``base_url``).
    3. Unified env vars: ``KUMIHO_LLM_API_KEY``, ``KUMIHO_LLM_PROVIDER``,
       ``KUMIHO_LLM_MODEL``, ``KUMIHO_LLM_BASE_URL``.
    4. Provider-specific env vars: ``OPENAI_API_KEY`` / ``ANTHROPIC_API_KEY``.
    5. Built-in defaults per provider.
    """

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        light_model: Optional[str] = None,
        base_url: Optional[str] = None,
        adapter: Optional[LLMAdapter] = None,
        client: Optional[Any] = None,
    ) -> None:
        # --- Fast path: caller provides a ready adapter ---
        if adapter is not None:
            self._adapter: Optional[LLMAdapter] = adapter
            self.provider = provider or "custom"
            self.model = model or "default"
            self.light_model = light_model or model or "default"
            self.api_key = api_key
            self._client = None
            self._base_url = base_url
            return

        # --- Resolve configuration (lightweight, no SDK init) ---
        resolved_key = api_key or os.getenv("KUMIHO_LLM_API_KEY")

        resolved_base_url = (
            base_url
            or os.getenv("KUMIHO_LLM_BASE_URL", "").strip()
            or None
        )

        resolved_provider = (
            (provider or "").strip().lower()
            or os.getenv("KUMIHO_LLM_PROVIDER", "").strip().lower()
            or self._provider_from_env_keys()
            or (_detect_provider(resolved_key) if resolved_key else "openai")
        )

        defaults = PROVIDER_DEFAULTS.get(resolved_provider, {})

        self.provider = resolved_provider
        self.api_key = resolved_key
        self.model = (
            model
            or os.getenv("KUMIHO_LLM_MODEL", "").strip()
            or defaults.get("model", "gpt-5.3-codex-spark" if "codex" in resolved_provider else "gpt-4o")
        )
        default_light_model = (
            self.model if "codex" in self.model.lower()
            else defaults.get("light_model", self.model)
        )
        self.light_model = (
            light_model
            or os.getenv("KUMIHO_LLM_LIGHT_MODEL", "").strip()
            or default_light_model
        )

        # --- Defer adapter construction until first use ---
        self._adapter = None
        self._client = client
        self._base_url = resolved_base_url

    @property
    def adapter(self) -> LLMAdapter:
        """Lazily build the LLM adapter on first use."""
        if self._adapter is not None:
            return self._adapter

        if self._client is not None:
            if self.provider == "anthropic":
                self._adapter = AnthropicAdapter(self._client)
            else:
                self._adapter = OpenAICompatAdapter(self._client)
        else:
            self._adapter = self._build_adapter(
                self.provider, self.api_key, self._base_url,
            )
        return self._adapter

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    async def summarize_conversation(
        self,
        messages: List[Dict[str, Any]],
        *,
        context: Optional[str] = None,
        strict: bool = False,
    ) -> Dict[str, Any]:
        """Summarize conversation into structured memory."""

        system_prompt = self._system_prompt()
        conversation_text = self._format_messages(messages)
        user_prompt = self._user_prompt(conversation_text, context=context)
        json_mode = build_summary_schema_mode()
        raw = ""

        try:
            raw = await self.adapter.chat(
                messages=[{"role": "user", "content": user_prompt}],
                model=self.model,
                system=system_prompt,
                max_tokens=2560,
                json_mode=json_mode,
            )
            result = self._parse_json(raw)
        except Exception as exc:
            diagnostics = self._build_diagnostics(
                model=self.model,
                json_mode=json_mode,
                raw_response=raw,
            )
            self._log_llm_failure(
                "summarize_conversation",
                error=exc,
                diagnostics=diagnostics,
            )
            if strict:
                raise
            return self._fallback_summary(messages, error=str(exc), debug=diagnostics)

        return self._normalize_summary(result, fallback_messages=messages)

    async def extract_topics(self, text: str) -> List[str]:
        """Extract topic keywords from text."""
        prompt = (
            "Extract 3-5 concise topic keywords from this text.\n\n"
            f"{text}\n\nTopics (comma-separated):"
        )
        response = await self.adapter.chat(
            messages=[{"role": "user", "content": prompt}],
            model=self.light_model,
            max_tokens=60,
        )
        topics = [item.strip().lower() for item in response.split(",") if item.strip()]
        return topics

    async def generate_implications(
        self,
        messages: List[Dict[str, Any]],
        *,
        context: Optional[str] = None,
    ) -> List[str]:
        """Generate prospective implications using the light model.

        Runs independently of summarization so it can be called in parallel.
        Returns 3-5 hypothetical future situations where this conversation
        would be the missing context, using *different* vocabulary than the
        original text to bridge semantic gaps in vector search.
        """
        conversation_text = self._format_messages(messages)
        prompt = (
            "Read this conversation and imagine someone months later says or "
            "does something that ONLY makes sense because of what happened here.\n"
            "Generate 3-5 short descriptions of those future situations.\n"
            "Rules:\n"
            "- Use DIFFERENT vocabulary than the original conversation\n"
            "- Focus on downstream behavioral changes, anxieties, preferences, "
            "or habits that would result from these events\n"
            "- Each description should be 1 sentence, max 20 words\n"
            '- Return ONLY a JSON object like {"implications": ["..."]}\n\n'
            f"Conversation:\n{conversation_text}\n\n"
            'JSON object with key "implications":'
        )

        raw = ""
        json_mode = build_string_array_wrapper_schema(
            "kumiho_implications_response",
            "implications",
        )
        try:
            raw = await self.adapter.chat(
                messages=[{"role": "user", "content": prompt}],
                model=self.light_model,
                max_tokens=512,
                json_mode=json_mode,
            )
            parsed = self._parse_json(raw)
            if isinstance(parsed, dict):
                implications = parsed.get("implications")
                if isinstance(implications, list):
                    return [str(item).strip() for item in implications if item][:5]
                for val in parsed.values():
                    if isinstance(val, list):
                        return [str(item).strip() for item in val if item][:5]
        except Exception as exc:
            self._log_llm_failure(
                "generate_implications",
                error=exc,
                diagnostics=self._build_diagnostics(
                    model=self.light_model,
                    json_mode=json_mode,
                    raw_response=raw,
                ),
            )

        return []

    # ------------------------------------------------------------------
    # Adapter construction
    # ------------------------------------------------------------------

    @staticmethod
    def _build_adapter(
        provider: str,
        api_key: Optional[str],
        base_url: Optional[str],
    ) -> LLMAdapter:
        resolved_base_url = (
            base_url
            or os.getenv("KUMIHO_LLM_BASE_URL", "").strip()
            or None
        )
        resolved_key = api_key or os.getenv("KUMIHO_LLM_API_KEY")

        if provider == "anthropic" and not resolved_base_url:
            key = resolved_key or os.getenv("ANTHROPIC_API_KEY")
            return AnthropicAdapter.create(api_key=key)

        # Default to OpenAI-compatible (covers openai, gemini, ollama, etc.)
        key = resolved_key or os.getenv("OPENAI_API_KEY")
        return OpenAICompatAdapter.create(api_key=key, base_url=resolved_base_url)

    @staticmethod
    def _provider_from_env_keys() -> str:
        """Detect provider from which provider-specific env var is set."""
        if os.getenv("OPENAI_API_KEY"):
            return "openai"
        if os.getenv("ANTHROPIC_API_KEY"):
            return "anthropic"
        return ""

    # ------------------------------------------------------------------
    # Prompt templates
    # ------------------------------------------------------------------

    @staticmethod
    def _system_prompt() -> str:
        return (
            "You are a memory extraction AI. Analyze conversations and extract structured knowledge.\n"
            "Output JSON in this exact schema:\n"
            "{\n"
            '  "type": "summary | fact | decision | action | reflection | error",\n'
            '  "title": "One-line summary (max 12 words)",\n'
            '  "summary": "Comprehensive summary (5-10 sentences) preserving ALL concrete details: dates, times, names, places, numbers, amounts, brands, model names, titles, roles, preferences, and outcomes",\n'
            '  "events": [\n'
            "    {\n"
            '      "event": "Specific incident or action that happened",\n'
            '      "when": "Date or time when it occurred (e.g. \'7 May 2023\', \'June 2023\', \'2022\', \'last week\') — use exact dates from the conversation when available",\n'
            '      "participants": ["person1"],\n'
            '      "consequence": "What changed as a result (behavioral change, decision, outcome)"\n'
            "    }\n"
            "  ],\n"
            '  "knowledge": {\n'
            '    "facts": [{"claim": "Specific factual claim with concrete detail", "certainty": "low | medium | high"}],\n'
            '    "decisions": [{"decision": "...", "reason": "..."}],\n'
            '    "actions": [{"task": "...", "status": "open | done | blocked"}],\n'
            '    "open_questions": ["..."]\n'
            "  },\n"
            '  "classification": {\n'
            '    "topics": ["topic1", "topic2"],\n'
            '    "entities": ["person1", "project1"]\n'
            "  }\n"
            "}\n"
            "Rules:\n"
            "- The summary should capture enough context that a future reader can answer specific factual questions without reading the full conversation\n"
            "- PRESERVE ALL DATES, TIMESTAMPS, AND TEMPORAL MARKERS mentioned in the conversation. Include them in the summary text and in each event's 'when' field. If a message is prefixed with a date like '[7 May 2023]', that is the date of the event.\n"
            "- Extract ALL specific events, incidents, and behavioral changes mentioned — even seemingly minor ones (accidents, purchases, lifestyle changes, health events, equipment failures)\n"
            "- Each event must be a concrete thing that happened, not a general topic or theme\n"
            "- Always include the consequence: what changed, what was decided, or what behavior resulted\n"
            "- Always include the 'when' field for each event — use exact dates from the conversation when available, otherwise use relative dates or 'unknown'\n"
            "- Extract ALL factual claims into knowledge.facts — include every piece of concrete information: names, places, brands, model numbers, job titles, relationships, hobbies, preferences, possessions, plans, amounts, measurements. If someone mentions owning a specific car, phone, or tool, that's a fact. If they mention where they work, live, or travel, that's a fact.\n"
            "- Extract decisions with their rationale into knowledge.decisions\n"
            "- Redact PII using placeholders like [EMAIL], [PHONE]\n"
            "- Focus on knowledge, not verbatim quotes"
        )

    @staticmethod
    def _format_messages(messages: List[Dict[str, Any]]) -> str:
        lines = []
        for msg in messages:
            role = str(msg.get("role", "unknown"))
            content = str(msg.get("content", ""))
            if content:
                lines.append(f"{role}: {content}")
        return "\n".join(lines)

    @staticmethod
    def _user_prompt(conversation_text: str, *, context: Optional[str]) -> str:
        context_line = f"\nContext: {context}\n" if context else ""
        return (
            "Conversation to summarize:\n\n"
            f"{conversation_text}\n"
            f"{context_line}\n"
            "Extract structured knowledge in JSON format."
        )

    def _build_diagnostics(
        self,
        *,
        model: str,
        json_mode: JsonMode,
        raw_response: str = "",
    ) -> Dict[str, Any]:
        return {
            "provider": self.provider,
            "model": model,
            "base_url": self._base_url or "",
            "json_mode": self._describe_json_mode(json_mode),
            "raw_response_len": len(raw_response or ""),
            "raw_response_preview": self._preview_text(raw_response),
        }

    def _log_llm_failure(
        self,
        operation: str,
        *,
        error: Exception,
        diagnostics: Dict[str, Any],
    ) -> None:
        logger.warning(
            "%s failed: %s | provider=%s model=%s base_url=%s json_mode=%s raw_len=%s raw_preview=%r",
            operation,
            error,
            diagnostics.get("provider", ""),
            diagnostics.get("model", ""),
            diagnostics.get("base_url", ""),
            diagnostics.get("json_mode", ""),
            diagnostics.get("raw_response_len", 0),
            diagnostics.get("raw_response_preview", ""),
        )

    @staticmethod
    def _describe_json_mode(json_mode: JsonMode) -> str:
        if isinstance(json_mode, dict):
            schema = json_mode.get("schema", {})
            properties = schema.get("properties", {})
            property_names = ",".join(sorted(str(key) for key in properties.keys()))
            return f"schema:{json_mode.get('name', '')}:type={schema.get('type', '')}:props={property_names}"
        if json_mode == "array":
            return "array"
        if json_mode:
            return "object"
        return "none"

    @staticmethod
    def _preview_text(text: str, *, limit: int = 240) -> str:
        compact = re.sub(r"\s+", " ", text or "").strip()
        if len(compact) <= limit:
            return compact
        return compact[: limit - 3] + "..."

    # ------------------------------------------------------------------
    # JSON parsing & fallbacks
    # ------------------------------------------------------------------

    @staticmethod
    def _parse_json(text: str) -> Any:
        candidates = []
        saw_json_like_input = False

        direct = (text or "").strip()
        if direct:
            candidates.append(direct)
            saw_json_like_input = direct.startswith(("```", "{", "["))

        normalized = MemorySummarizer._strip_markdown_code_fences(direct)
        if normalized and normalized not in candidates:
            candidates.append(normalized)
            saw_json_like_input = saw_json_like_input or normalized.startswith(("{", "["))

        for candidate in list(candidates):
            extracted = MemorySummarizer._extract_balanced_json(candidate)
            if extracted and extracted not in candidates:
                candidates.append(extracted)
                saw_json_like_input = True

        last_error: Optional[Exception] = None
        for candidate in candidates:
            try:
                return json.loads(candidate)
            except json.JSONDecodeError as exc:
                last_error = exc

        if saw_json_like_input and last_error is not None:
            raise last_error
        raise ValueError("No valid JSON found in summarizer response")

    @staticmethod
    def _strip_markdown_code_fences(text: str) -> str:
        stripped = (text or "").strip()
        if not stripped.startswith("```"):
            return stripped

        stripped = re.sub(r"^```(?:json)?\s*\n?", "", stripped, flags=re.IGNORECASE)
        stripped = re.sub(r"\n?```\s*$", "", stripped)
        return stripped.strip()

    @staticmethod
    def _extract_balanced_json(text: str) -> str:
        source = (text or "").strip()
        if not source:
            return ""

        start = -1
        opening = ""
        closing = ""
        for idx, ch in enumerate(source):
            if ch == "{":
                start = idx
                opening = "{"
                closing = "}"
                break
            if ch == "[":
                start = idx
                opening = "["
                closing = "]"
                break

        if start < 0:
            return ""

        depth = 0
        in_string = False
        escaped = False
        for idx in range(start, len(source)):
            ch = source[idx]
            if in_string:
                if escaped:
                    escaped = False
                elif ch == "\\":
                    escaped = True
                elif ch == '"':
                    in_string = False
                continue

            if ch == '"':
                in_string = True
                continue

            if ch == opening:
                depth += 1
            elif ch == closing:
                depth -= 1
                if depth == 0:
                    return source[start : idx + 1]

        return ""

    @staticmethod
    def _normalize_summary(result: Dict[str, Any], *, fallback_messages: List[Dict[str, Any]]) -> Dict[str, Any]:
        summary = dict(result)
        summary.setdefault("type", "summary")
        summary.setdefault("title", "Conversation summary")
        raw_summary = summary.get("summary")
        if isinstance(raw_summary, str) and raw_summary.strip():
            summary["summary"] = raw_summary.strip()
        else:
            summary["summary"] = MemorySummarizer._fallback_summary_text(fallback_messages)
            summary.setdefault(
                "error",
                "Summarizer response did not include a non-empty summary",
            )
        summary.setdefault("knowledge", {})
        summary.setdefault("classification", {})
        summary.setdefault("events", [])
        summary.setdefault("implications", [])
        summary["knowledge"].setdefault("facts", [])
        summary["knowledge"].setdefault("decisions", [])
        summary["knowledge"].setdefault("actions", [])
        summary["knowledge"].setdefault("open_questions", [])
        summary["classification"].setdefault("topics", [])
        summary["classification"].setdefault("entities", [])
        return summary

    @staticmethod
    def _fallback_summary(
        messages: List[Dict[str, Any]],
        *,
        error: str,
        debug: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        text = MemorySummarizer._fallback_summary_text(messages)
        result = {
            "type": "summary",
            "title": "Conversation summary",
            "summary": text,
            "events": [],
            "implications": [],
            "knowledge": {"facts": [], "decisions": [], "actions": [], "open_questions": []},
            "classification": {"topics": [], "entities": []},
            "error": error,
        }
        if debug:
            result["debug"] = debug
        return result

    @staticmethod
    def _fallback_summary_text(messages: List[Dict[str, Any]]) -> str:
        if not messages:
            return "No conversation content available."

        def _strip_injected_blocks(text: str) -> str:
            cleaned = text
            for tag in ("kumiho_memory", "kumiho_project", "kumiho_instructions"):
                cleaned = re.sub(
                    rf"<{tag}>.*?</{tag}>",
                    "",
                    cleaned,
                    flags=re.DOTALL | re.IGNORECASE,
                )
            return cleaned.strip()

        preferred_roles = {"user", "assistant"}
        for msg in reversed(messages):
            role = str(msg.get("role", "")).strip().lower()
            if role and role not in preferred_roles:
                continue
            snippet = _strip_injected_blocks(str(msg.get("content") or ""))
            if len(snippet) > 500:
                snippet = f"{snippet[:497]}..."
            if snippet:
                return snippet

        return "Conversation summary unavailable."



"""CLI entry point for kumiho-memory.

Usage::

    # Run Dream State (reads config from ~/.kumiho/preferences.json)
    python -m kumiho_memory dream

    # Dry run (assess but don't mutate)
    python -m kumiho_memory dream --dry-run

    # Custom project and batch size
    python -m kumiho_memory dream --project MyProject --batch-size 10

    # Via the installed console script
    kumiho-memory dream
"""

from __future__ import annotations

import argparse
import asyncio
import json
import logging
import os
import sys
from pathlib import Path

logger = logging.getLogger("kumiho_memory")


def _load_preferences() -> dict:
    """Load ~/.kumiho/preferences.json if it exists."""
    prefs_path = Path.home() / ".kumiho" / "preferences.json"
    if prefs_path.exists():
        try:
            return json.loads(prefs_path.read_text("utf-8"))
        except (json.JSONDecodeError, OSError):
            pass
    return {}


def _configure_llm_from_prefs(prefs: dict, section: str = "dreamState") -> None:
    """Set KUMIHO_LLM_* env vars from preferences.json if not already set.

    The setup wizard stores shared direct-provider config plus per-feature models::

        {
            "llm": { "provider": "gemini", "apiKey": "...", "baseUrl": "..." },
            "dreamState": { "model": { "provider": "gemini", "model": "..." } }
        }

    We merge the shared ``llm`` block first, then let the section-specific
    model config override it before translating to the env vars that
    MemorySummarizer reads.
    """
    model_cfg = {
        **prefs.get("llm", {}),
        **prefs.get(section, {}).get("model", {}),
    }
    if not model_cfg:
        return

    mapping = {
        "provider": "KUMIHO_LLM_PROVIDER",
        "model": "KUMIHO_LLM_MODEL",
        "apiKey": "KUMIHO_LLM_API_KEY",
        "baseUrl": "KUMIHO_LLM_BASE_URL",
    }
    for key, env_var in mapping.items():
        val = model_cfg.get(key)
        if val and not os.environ.get(env_var):
            os.environ[env_var] = val


def cmd_dream(args: argparse.Namespace) -> int:
    """Run a Dream State consolidation cycle."""
    from kumiho_memory import DreamState

    prefs = _load_preferences()
    _configure_llm_from_prefs(prefs, "dreamState")

    ds = DreamState(
        project=args.project,
        batch_size=args.batch_size,
        dry_run=args.dry_run,
        max_deprecation_ratio=args.max_deprecation_ratio,
        allow_published_deprecation=args.allow_published_deprecation,
    )

    result = asyncio.run(ds.run())

    # Print summary
    print(json.dumps(result, indent=2, default=str))
    errors = result.get("errors", [])
    return 1 if errors else 0


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(
        prog="kumiho-memory",
        description="Kumiho Memory CLI - standalone memory operations",
    )
    parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="Enable debug logging",
    )

    sub = parser.add_subparsers(dest="command")

    # -- dream subcommand --
    dream = sub.add_parser(
        "dream",
        help="Run Dream State memory consolidation",
        description="Run a Dream State cycle: assess recent memories, "
        "deprecate stale ones, enrich metadata, create edges.",
    )
    dream.add_argument(
        "--project",
        default="CognitiveMemory",
        help="Kumiho project name (default: CognitiveMemory)",
    )
    dream.add_argument(
        "--batch-size",
        type=int,
        default=20,
        help="Memories per LLM assessment batch (default: 20)",
    )
    dream.add_argument(
        "--dry-run",
        action="store_true",
        help="Assess but do not mutate anything",
    )
    dream.add_argument(
        "--max-deprecation-ratio",
        type=float,
        default=0.5,
        help="Max fraction of memories to deprecate per run (default: 0.5)",
    )
    dream.add_argument(
        "--allow-published-deprecation",
        action="store_true",
        help="Allow deprecation of published items (use with caution)",
    )

    parsed = parser.parse_args(argv)

    # Configure logging
    logging.basicConfig(
        level=logging.DEBUG if parsed.verbose else logging.INFO,
        format="%(asctime)s %(levelname)s [%(name)s] %(message)s",
        datefmt="%H:%M:%S",
    )

    if parsed.command == "dream":
        return cmd_dream(parsed)

    parser.print_help()
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Dream State — scheduled memory consolidation processor.

The Dream State runs periodically (e.g. nightly at 3 AM) to:

1. Query revisions created or updated since the last run.
2. Fetch full revision data for changed memories.
3. Inspect bundles for new conversation groupings.
4. Use an LLM to assess each memory: deprecate low-value ones,
   enrich metadata / tags, and suggest relationships.
5. Apply the assessed changes to the Kumiho graph.
6. Persist the timestamp and generate a Markdown report.

Usage::

    from kumiho_memory import DreamState

    ds = DreamState(project="CognitiveMemory")
    report = await ds.run()
    print(report)
"""

from __future__ import annotations

import asyncio
import json
import logging
import re
import time
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

from kumiho_memory.summarization import (
    LLMAdapter,
    MemorySummarizer,
    _json_schema_mode,
    _strict_object_schema,
)

logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------
# Data classes
# ---------------------------------------------------------------------------


@dataclass
class MemoryAssessment:
    """LLM-produced assessment for a single memory revision."""

    revision_kref: str
    relevance_score: float
    should_deprecate: bool
    deprecation_reason: str = ""
    suggested_tags: List[str] = field(default_factory=list)
    metadata_updates: Dict[str, str] = field(default_factory=dict)
    related_memories: List[Tuple[str, str]] = field(default_factory=list)
    """List of ``(target_revision_kref, edge_type)`` tuples."""


@dataclass
class DreamStateStats:
    """Counters accumulated during a single Dream State run."""

    events_processed: int = 0
    revisions_assessed: int = 0
    deprecated: int = 0
    metadata_updated: int = 0
    tags_added: int = 0
    edges_created: int = 0
    last_cursor: Optional[str] = None  # Kept for backward-compat in report dict
    duration_ms: int = 0
    errors: List[str] = field(default_factory=list)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

_ASSESSMENT_SYSTEM_PROMPT = """\
You are a memory consolidation agent performing "Dream State" processing.
You will receive an array of memories (each with an index, title, summary,
type, tags, and metadata). Return a JSON object with a single key
``assessments`` whose value is an array of assessment objects. For **each**
memory include the following fields:

1. index (int): The memory's index in the input array.
2. relevance_score (float 0.0-1.0): How useful is this memory for future
   interactions?
3. should_deprecate (bool): True if the memory should be deprecated.
4. deprecation_reason (str): Why (empty string if keeping).
5. suggested_tags (List[str]): Additional tags for better retrieval.
6. metadata_updates (List[{"key": str, "value": str}]): Metadata key/value
   corrections or enrichments. Return ``[]`` if none.
7. related_indices (List[int]): Indices of related memories in THIS batch.
8. relationship_type (str): Edge type for related memories — one of
   DERIVED_FROM, REFERENCED, DEPENDS_ON, SUPERSEDES.  Empty string if none.

Return ONLY a JSON object like:
{"assessments": [ ... ]}.

Guidelines:
- Be conservative: when in doubt, KEEP the memory.
- Deprecate ONLY if the memory is: a near-duplicate of another memory in
  this batch, clearly superseded by newer information, trivially obvious,
  or contains no actionable information.
- Tags should aid retrieval: topic keywords, action types, entity names,
  project identifiers.
- Suggest relationships for memories that reference the same topic,
  project, or decision chain.
"""

_ASSESSMENT_SCHEMA_MODE = _json_schema_mode(
    "kumiho_assessments_response",
    _strict_object_schema({
        "assessments": {
            "type": "array",
            "items": _strict_object_schema({
                "index": {"type": "integer"},
                "relevance_score": {"type": "number"},
                "should_deprecate": {"type": "boolean"},
                "deprecation_reason": {"type": "string"},
                "suggested_tags": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "metadata_updates": {
                    "type": "array",
                    "items": _strict_object_schema({
                        "key": {"type": "string"},
                        "value": {"type": "string"},
                    }),
                },
                "related_indices": {
                    "type": "array",
                    "items": {"type": "integer"},
                },
                "relationship_type": {"type": "string"},
            }),
        },
    }),
)


def _parse_assessments(raw: str) -> List[Dict[str, Any]]:
    """Best-effort parse of LLM JSON output."""
    # Try direct parse first
    try:
        parsed = json.loads(raw)
        if isinstance(parsed, list):
            return parsed
        if isinstance(parsed, dict) and "assessments" in parsed:
            return parsed["assessments"]
        return [parsed]
    except json.JSONDecodeError:
        pass

    # Try to extract a JSON array from markdown fences
    match = re.search(r"```(?:json)?\s*(\[.*?\])\s*```", raw, re.DOTALL)
    if match:
        try:
            return json.loads(match.group(1))
        except json.JSONDecodeError:
            pass

    # Last resort — look for bare array
    match = re.search(r"\[.*\]", raw, re.DOTALL)
    if match:
        try:
            return json.loads(match.group(0))
        except json.JSONDecodeError:
            pass

    return []


# ---------------------------------------------------------------------------
# DreamState
# ---------------------------------------------------------------------------


class DreamState:
    """Scheduled memory consolidation processor.

    Parameters
    ----------
    project:
        Kumiho project name (default ``CognitiveMemory``).
    summarizer:
        Existing :class:`MemorySummarizer` to reuse for LLM calls.
    llm_adapter:
        Raw :class:`LLMAdapter` — a ``MemorySummarizer`` is built around it.
    artifact_root:
        Local directory for writing report artifacts.
    cursor_item_name:
        Item name used to persist the run timestamp and
        Dream State reports (default ``_dream_state``).
    batch_size:
        Number of memories to assess per LLM call.
    dry_run:
        If *True*, assess but do **not** mutate anything in Kumiho.
    max_deprecation_ratio:
        Maximum fraction of memories that may be deprecated per run.
        Must be between 0.1 and 0.9 (default 0.5).
    allow_published_deprecation:
        If *True*, published items may be deprecated. Use with caution.
        When relaxed, a warning is logged and recorded in the audit report.
    kind_filter:
        Item kind to process (default ``conversation``).  Set to empty
        string to process all item kinds.
    """

    def __init__(
        self,
        *,
        project: str = "CognitiveMemory",
        summarizer: Optional[MemorySummarizer] = None,
        llm_adapter: Optional[LLMAdapter] = None,
        artifact_root: Optional[str] = None,
        cursor_item_name: str = "_dream_state",
        batch_size: int = 20,
        dry_run: bool = False,
        max_deprecation_ratio: float = 0.5,
        allow_published_deprecation: bool = False,
        kind_filter: str = "conversation",
        # Legacy parameters — accepted but ignored for backward compatibility
        routing_key_filter: str = "revision.*",
        event_timeout: float = 10.0,
    ) -> None:
        self.project = project
        self.cursor_item_name = cursor_item_name
        self.batch_size = batch_size
        self.kind_filter = kind_filter
        self.dry_run = dry_run

        if not (0.1 <= max_deprecation_ratio <= 0.9):
            raise ValueError(
                f"max_deprecation_ratio must be between 0.1 and 0.9, "
                f"got {max_deprecation_ratio}"
            )
        self.max_deprecation_ratio = max_deprecation_ratio
        self.allow_published_deprecation = allow_published_deprecation

        import os

        self.artifact_root = artifact_root or os.getenv(
            "KUMIHO_MEMORY_ARTIFACT_ROOT",
            os.path.join(os.path.expanduser("~"), ".kumiho", "artifacts"),
        )
        self.space_page_size = max(
            1,
            int(os.getenv("KUMIHO_DREAM_STATE_SPACE_PAGE_SIZE", "100")),
        )
        self.item_page_size = max(
            1,
            int(os.getenv("KUMIHO_DREAM_STATE_ITEM_PAGE_SIZE", "100")),
        )

        if summarizer is not None:
            self.summarizer = summarizer
        elif llm_adapter is not None:
            self.summarizer = MemorySummarizer(adapter=llm_adapter)
        else:
            self.summarizer = MemorySummarizer()

        # Will be resolved lazily on first run.
        self._cursor_item_kref: Optional[str] = None

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    async def run(self) -> Dict[str, Any]:
        """Execute a full Dream State cycle.

        Returns a report dict with counters and the final cursor.
        """
        import kumiho

        start = time.monotonic()
        stats = DreamStateStats()
        run_started_at = datetime.now(timezone.utc).isoformat()

        try:
            # 1. Ensure cursor item exists
            cursor_kref = self._ensure_cursor_item(kumiho)

            # 2. Load last_run_at timestamp
            last_run_at = self._load_last_run_at(kumiho, cursor_kref)

            # 3. Collect revisions created/updated since last run
            revisions = await asyncio.to_thread(
                self._collect_revisions, kumiho, last_run_at
            )
            stats.events_processed = len(revisions)
            if not revisions:
                logger.info(
                    "Dream State: no new revisions since %s",
                    last_run_at or "beginning",
                )
                stats.duration_ms = int((time.monotonic() - start) * 1000)
                # Still save timestamp so next run skips this window
                self._save_last_run_at(kumiho, cursor_kref, run_started_at)
                return self._build_result(stats, report_kref=None)

            # 4. Inspect bundles (from revision item krefs)
            bundle_context = self._inspect_bundles_from_revisions(
                kumiho, revisions
            )

            # 5. Assess in batches
            all_assessments: List[MemoryAssessment] = []
            for i in range(0, len(revisions), self.batch_size):
                batch = revisions[i : i + self.batch_size]
                assessments = await self._assess_batch(batch, bundle_context)
                all_assessments.extend(assessments)

            stats.revisions_assessed = len(all_assessments)

            # 6. Apply actions
            self._apply_actions(kumiho, all_assessments, stats)

            # 7. Save last_run_at
            self._save_last_run_at(kumiho, cursor_kref, run_started_at)

            # 8. Generate report
            stats.duration_ms = int((time.monotonic() - start) * 1000)
            report_kref = self._generate_report(
                kumiho, cursor_kref, stats, all_assessments
            )

            return self._build_result(stats, report_kref=report_kref)

        except Exception as exc:
            stats.errors.append(str(exc))
            stats.duration_ms = int((time.monotonic() - start) * 1000)
            logger.exception("Dream State run failed")
            return {
                "success": False,
                "error": str(exc),
                **self._stats_dict(stats),
            }

    # ------------------------------------------------------------------
    # Timestamp management (replaces event-stream cursor)
    # ------------------------------------------------------------------

    def _ensure_cursor_item(self, sdk: Any) -> str:
        """Return the kref of the ``_dream_state`` item, creating it if
        necessary."""
        if self._cursor_item_kref is not None:
            return self._cursor_item_kref

        kref_uri = f"kref://{self.project}/{self.cursor_item_name}.conversation"
        try:
            item = sdk.get_item(kref_uri)
            if item is not None:
                self._cursor_item_kref = item.kref.uri
                return self._cursor_item_kref
        except Exception:
            pass

        # Create the item — first ensure parent space exists.
        try:
            project = sdk.get_project(self.project)
            if project is None:
                raise RuntimeError(
                    f"Project '{self.project}' does not exist"
                )

            try:
                space = project.get_space(self.cursor_item_name)
            except Exception:
                space = None
            if space is None:
                space = project.create_space(self.cursor_item_name)

            item = space.create_item(self.cursor_item_name, "conversation")
            self._cursor_item_kref = item.kref.uri
        except Exception:
            # Fallback: item might already exist (race)
            try:
                item = sdk.get_item(kref_uri)
                self._cursor_item_kref = item.kref.uri
            except Exception as inner:
                raise RuntimeError(
                    f"Failed to ensure cursor item: {inner}"
                ) from inner

        return self._cursor_item_kref  # type: ignore[return-value]

    def _load_last_run_at(self, sdk: Any, cursor_kref: str) -> Optional[str]:
        """Read the last-saved run timestamp (ISO format).

        Tries the gRPC attribute first, then falls back to the local
        cursor file written by ``_save_cursor_local``.
        """
        try:
            value = sdk.get_attribute(cursor_kref, "last_run_at")
            if value:
                return value
        except Exception:
            pass

        # Fall back to local cursor file
        return self._load_cursor_local()

    # ------------------------------------------------------------------
    # Local cursor file (fallback when gRPC is unavailable)
    # ------------------------------------------------------------------

    @property
    def _cursor_file(self) -> Path:
        return (
            Path(self.artifact_root)
            / self.project
            / self.cursor_item_name
            / "cursor.json"
        )

    def _save_cursor_local(self, run_at: str) -> None:
        """Write the cursor timestamp to a local JSON file."""
        try:
            self._cursor_file.parent.mkdir(parents=True, exist_ok=True)
            self._cursor_file.write_text(
                json.dumps({"last_run_at": run_at}), encoding="utf-8"
            )
        except Exception as exc:
            logger.warning("Failed to write local cursor file: %s", exc)

    def _load_cursor_local(self) -> Optional[str]:
        """Read the cursor timestamp from the local JSON file."""
        try:
            if self._cursor_file.exists():
                data = json.loads(
                    self._cursor_file.read_text(encoding="utf-8")
                )
                return data.get("last_run_at")
        except Exception:
            pass
        return None

    def _save_last_run_at(
        self, sdk: Any, cursor_kref: str, run_at: str
    ) -> None:
        """Persist the run timestamp.

        The kumiho SDK client includes a ``_TransientRetryInterceptor``
        that automatically retries on UNAVAILABLE / DEADLINE_EXCEEDED
        with exponential backoff.  If the call still fails after SDK
        retries, we fall back to a local cursor file so the next run
        can pick up where this one left off.
        """
        try:
            sdk.set_attribute(cursor_kref, "last_run_at", run_at)
            # Also persist locally as a safety net
            self._save_cursor_local(run_at)
        except Exception as exc:
            logger.error(
                "Failed to save last_run_at via gRPC: %s. "
                "Falling back to local cursor file.",
                exc,
            )
            self._save_cursor_local(run_at)

    # ------------------------------------------------------------------
    # Revision collection (replaces event stream)
    # ------------------------------------------------------------------

    def _list_project_spaces(self, project: Any) -> List[Any]:
        """Enumerate project spaces without relying on one recursive RPC."""
        root_path = f"/{self.project}"
        discovered: List[Any] = []
        seen_paths = set()
        pending_paths = [root_path]

        while pending_paths:
            parent_path = pending_paths.pop(0)
            cursor: Optional[str] = None

            while True:
                try:
                    page = project.get_spaces(
                        parent_path=parent_path,
                        recursive=False,
                        page_size=self.space_page_size,
                        cursor=cursor,
                    )
                except TypeError:
                    # Older SDK stubs/tests only support the legacy recursive API.
                    spaces = list(project.get_spaces(recursive=True))
                    logger.info(
                        "Dream State: using legacy recursive space enumeration "
                        "for project %s",
                        self.project,
                    )
                    return spaces
                except Exception as exc:
                    raise RuntimeError(
                        "Failed to list child spaces under "
                        f"'{parent_path}' (cursor={cursor or '-'})"
                    ) from exc

                children = list(page)
                for space in children:
                    path = getattr(space, "path", "")
                    if not path or path in seen_paths:
                        continue
                    seen_paths.add(path)
                    discovered.append(space)
                    pending_paths.append(path)

                cursor = getattr(page, "next_cursor", None)
                if not cursor:
                    break

        return discovered

    def _list_space_items(self, sdk: Any, space_path: str) -> List[Any]:
        """List items in a space in bounded pages to avoid RPC deadlines."""
        client = sdk.get_client()
        kind_arg = self.kind_filter if self.kind_filter else ""
        collected: List[Any] = []
        cursor: Optional[str] = None

        while True:
            try:
                page = client.get_items(
                    parent_path=space_path,
                    kind_filter=kind_arg,
                    page_size=self.item_page_size,
                    cursor=cursor,
                    include_deprecated=False,
                )
            except TypeError:
                page = client.get_items(
                    parent_path=space_path,
                    kind_filter=kind_arg,
                    include_deprecated=False,
                )
            except Exception as exc:
                raise RuntimeError(
                    "Failed to list items in "
                    f"'{space_path}' (cursor={cursor or '-'})"
                ) from exc

            collected.extend(list(page))

            cursor = getattr(page, "next_cursor", None)
            if not cursor:
                break

        return collected

    def _collect_revisions(
        self, sdk: Any, last_run_at: Optional[str]
    ) -> list:
        """Enumerate all spaces in the project, list items, and collect
        latest revisions that were created after *last_run_at*.

        This replaces the old event-stream approach which suffered from
        gRPC DEADLINE_EXCEEDED errors and cursor issues.  Direct revision
        queries are reliable and catch both new items and stacked revisions
        on existing items.
        """
        try:
            project = sdk.get_project(self.project)
            if project is None:
                logger.warning("Project '%s' not found", self.project)
                return []
        except Exception as exc:
            logger.warning("Failed to get project '%s': %s", self.project, exc)
            return []

        # Parse the cutoff timestamp
        cutoff: Optional[datetime] = None
        if last_run_at:
            try:
                cutoff = datetime.fromisoformat(last_run_at)
                # Ensure timezone-aware
                if cutoff.tzinfo is None:
                    cutoff = cutoff.replace(tzinfo=timezone.utc)
            except (ValueError, TypeError):
                logger.warning(
                    "Invalid last_run_at timestamp '%s', processing all",
                    last_run_at,
                )

        # Enumerate spaces breadth-first in small pages so large projects do
        # not rely on a single recursive GetChildSpaces RPC.
        try:
            spaces = self._list_project_spaces(project)
        except Exception as exc:
            logger.warning("Failed to enumerate spaces: %s", exc)
            return []

        collected: list = []
        cursor_item_kref = self._cursor_item_kref
        space_paths = [f"/{self.project}"]
        seen_space_paths = {f"/{self.project}"}
        for space in spaces:
            path = getattr(space, "path", "")
            if not path or path in seen_space_paths:
                continue
            seen_space_paths.add(path)
            space_paths.append(path)

        for space_path in space_paths:
            try:
                items = self._list_space_items(sdk, space_path)
            except Exception as exc:
                logger.warning(
                    "Failed to list items in space '%s': %s",
                    space_path, exc,
                )
                continue

            for item in items:
                # Skip the _dream_state cursor item itself
                item_kref = item.kref.uri if hasattr(item, "kref") else ""
                if cursor_item_kref and item_kref == cursor_item_kref:
                    continue

                try:
                    # Get the latest revision
                    rev = item.get_revision_by_tag("latest")
                    if rev is None:
                        continue
                except Exception:
                    # No 'latest' tag — try getting all revisions
                    try:
                        revs = item.get_revisions()
                        if not revs:
                            continue
                        rev = revs[-1]  # Most recent
                    except Exception:
                        continue

                # Skip deprecated revisions
                if getattr(rev, "deprecated", False):
                    continue

                # Filter by created_at timestamp
                if cutoff is not None and rev.created_at:
                    try:
                        rev_time = datetime.fromisoformat(rev.created_at)
                        if rev_time.tzinfo is None:
                            rev_time = rev_time.replace(tzinfo=timezone.utc)
                        if rev_time <= cutoff:
                            continue
                    except (ValueError, TypeError):
                        pass  # Can't parse — include it to be safe

                collected.append(rev)

        logger.info(
            "Dream State: collected %d revisions since %s",
            len(collected),
            last_run_at or "beginning",
        )
        return collected

    # ------------------------------------------------------------------
    # Bundle inspection
    # ------------------------------------------------------------------

    def _inspect_bundles_from_revisions(
        self, sdk: Any, revisions: list
    ) -> Dict[str, list]:
        """For any bundle items among the collected revisions, fetch members."""
        bundles: Dict[str, list] = {}
        for rev in revisions:
            item_kref = ""
            try:
                item_kref = rev.item_kref.uri if hasattr(rev, "item_kref") else ""
            except Exception:
                continue
            if ".bundle" not in item_kref:
                continue
            if item_kref in bundles:
                continue
            try:
                bundle = sdk.get_item(item_kref)
                if bundle is not None and hasattr(bundle, "get_members"):
                    bundles[item_kref] = bundle.get_members()
            except Exception as exc:
                logger.warning("Failed to inspect bundle %s: %s", item_kref, exc)

        return bundles

    # ------------------------------------------------------------------
    # LLM assessment
    # ------------------------------------------------------------------

    async def _assess_batch(
        self,
        revisions: list,
        bundle_context: Dict[str, list],
    ) -> List[MemoryAssessment]:
        """Send a batch of revisions to the LLM for assessment."""
        if not revisions:
            return []

        # Build the user prompt — serialise each revision to JSON-like text.
        memories: List[Dict[str, Any]] = []
        kref_by_index: Dict[int, str] = {}

        for idx, rev in enumerate(revisions):
            meta = dict(getattr(rev, "metadata", {}) or {})
            entry: Dict[str, Any] = {
                "index": idx,
                "kref": rev.kref.uri if hasattr(rev, "kref") else str(rev),
                "title": meta.get("title", ""),
                "summary": meta.get("summary", ""),
                "type": meta.get("type", meta.get("memory_type", "")),
                "tags": meta.get("tags", ""),
                "topics": meta.get("topics", ""),
            }
            kref_by_index[idx] = entry["kref"]
            memories.append(entry)

        # Include bundle context if available.
        bundle_info = ""
        if bundle_context:
            parts = []
            for bkref, members in bundle_context.items():
                member_strs = []
                for m in members:
                    mkref = m.item_kref.uri if hasattr(m, "item_kref") else str(m)
                    member_strs.append(mkref)
                parts.append(f"Bundle {bkref}: members={member_strs}")
            bundle_info = "\n\nBundle groupings:\n" + "\n".join(parts)

        user_prompt = (
            "Assess the following memories:\n\n"
            + json.dumps(memories, indent=2, default=str)
            + bundle_info
        )

        try:
            raw = await self.summarizer.adapter.chat(
                messages=[{"role": "user", "content": user_prompt}],
                model=self.summarizer.model,
                system=_ASSESSMENT_SYSTEM_PROMPT,
                max_tokens=2048,
                json_mode=_ASSESSMENT_SCHEMA_MODE,
            )
        except Exception as exc:
            logger.warning("LLM assessment failed: %s", exc)
            return []

        parsed = _parse_assessments(raw)

        # Convert to MemoryAssessment objects.
        assessments: List[MemoryAssessment] = []
        for item in parsed:
            idx = item.get("index", -1)
            rev_kref = kref_by_index.get(idx, "")
            if not rev_kref:
                continue

            related: List[Tuple[str, str]] = []
            rel_type = item.get("relationship_type", "")
            for rel_idx in item.get("related_indices", []):
                target = kref_by_index.get(rel_idx)
                if target and target != rev_kref:
                    related.append((target, rel_type or "REFERENCED"))

            raw_metadata_updates = item.get("metadata_updates", {})
            metadata_updates: Dict[str, str] = {}
            if isinstance(raw_metadata_updates, dict):
                metadata_updates = {
                    str(key): str(value)
                    for key, value in raw_metadata_updates.items()
                    if key and value is not None
                }
            elif isinstance(raw_metadata_updates, list):
                metadata_updates = {
                    str(entry.get("key")): str(entry.get("value"))
                    for entry in raw_metadata_updates
                    if isinstance(entry, dict) and entry.get("key") and entry.get("value") is not None
                }

            assessments.append(
                MemoryAssessment(
                    revision_kref=rev_kref,
                    relevance_score=float(item.get("relevance_score", 0.5)),
                    should_deprecate=bool(item.get("should_deprecate", False)),
                    deprecation_reason=item.get("deprecation_reason", ""),
                    suggested_tags=list(item.get("suggested_tags", [])),
                    metadata_updates=metadata_updates,
                    related_memories=related,
                )
            )

        return assessments

    # ------------------------------------------------------------------
    # Apply actions
    # ------------------------------------------------------------------

    def _apply_actions(
        self,
        sdk: Any,
        assessments: List[MemoryAssessment],
        stats: DreamStateStats,
    ) -> None:
        """Apply the LLM-recommended changes to the Kumiho graph."""
        if self.dry_run:
            logger.info("Dry run — skipping %d actions", len(assessments))
            return

        if not assessments:
            return

        client = sdk.get_client()

        # Safety: cap deprecation per run (spec §9.4.4).
        deprecation_limit = max(1, int(len(assessments) * self.max_deprecation_ratio))
        deprecations_done = 0

        for assessment in assessments:
            kref_str = assessment.revision_kref
            try:
                kref = sdk.Kref(kref_str)
            except Exception:
                stats.errors.append(f"Invalid kref: {kref_str}")
                continue

            # --- Deprecate ---
            if assessment.should_deprecate:
                try:
                    is_published = client.has_tag(kref, "published")
                    if is_published and not self.allow_published_deprecation:
                        logger.info(
                            "Skipping deprecation of published revision %s",
                            kref_str,
                        )
                    elif deprecations_done >= deprecation_limit:
                        logger.info(
                            "Deprecation limit reached (%d/%d), skipping %s",
                            deprecations_done,
                            deprecation_limit,
                            kref_str,
                        )
                    else:
                        if is_published:
                            logger.warning(
                                "Published protection RELAXED — deprecating published revision %s",
                                kref_str,
                            )
                        client.set_deprecated(kref, True)
                        stats.deprecated += 1
                        deprecations_done += 1
                except Exception as exc:
                    stats.errors.append(f"deprecate {kref_str}: {exc}")

            # --- Tags ---
            for tag in assessment.suggested_tags:
                try:
                    client.tag_revision(kref, tag)
                    stats.tags_added += 1
                except Exception as exc:
                    stats.errors.append(f"tag {kref_str} '{tag}': {exc}")

            # --- Metadata updates ---
            if assessment.metadata_updates:
                try:
                    client.update_revision_metadata(
                        kref, assessment.metadata_updates
                    )
                    stats.metadata_updated += 1
                except Exception as exc:
                    stats.errors.append(f"metadata {kref_str}: {exc}")

            # --- Relationships / edges ---
            for target_kref_str, edge_type in assessment.related_memories:
                try:
                    target_kref = sdk.Kref(target_kref_str)
                    # create_edge needs Revision objects; fetch them.
                    source_rev = sdk.get_revision(kref_str)
                    target_rev = sdk.get_revision(target_kref_str)
                    if source_rev and target_rev:
                        client.create_edge(
                            source_rev, target_rev, edge_type
                        )
                        stats.edges_created += 1
                except Exception as exc:
                    stats.errors.append(
                        f"edge {kref_str} → {target_kref_str}: {exc}"
                    )

    # ------------------------------------------------------------------
    # Report generation
    # ------------------------------------------------------------------

    def _generate_report(
        self,
        sdk: Any,
        cursor_kref: str,
        stats: DreamStateStats,
        assessments: List[MemoryAssessment],
    ) -> Optional[str]:
        """Create a report revision + artifact on the cursor item."""
        now_iso = datetime.now(timezone.utc).isoformat()
        markdown = self._build_report_markdown(
            stats, assessments, now_iso,
            allow_published_deprecation=self.allow_published_deprecation,
        )

        # Write artifact to local storage.
        safe_ts = now_iso.replace(":", "").replace("-", "").split(".")[0]
        artifact_dir = (
            Path(self.artifact_root)
            / self.project
            / self.cursor_item_name
            / "reports"
        )
        artifact_dir.mkdir(parents=True, exist_ok=True)
        artifact_path = artifact_dir / f"dream_state_{safe_ts}.md"
        artifact_path.write_text(markdown, encoding="utf-8")

        # Create revision with metadata.
        try:
            item = sdk.get_item(cursor_kref)
            if item is None:
                return None

            revision = item.create_revision(
                metadata={
                    "type": "dream_state_report",
                    "events_processed": str(stats.events_processed),
                    "revisions_assessed": str(stats.revisions_assessed),
                    "deprecated": str(stats.deprecated),
                    "metadata_updated": str(stats.metadata_updated),
                    "tags_added": str(stats.tags_added),
                    "edges_created": str(stats.edges_created),
                    "cursor": stats.last_cursor or "",
                    "run_at": now_iso,
                    "duration_ms": str(stats.duration_ms),
                },
            )
            revision.create_artifact("report", str(artifact_path))
            return revision.kref.uri
        except Exception as exc:
            logger.warning("Failed to create report revision: %s", exc)
            stats.errors.append(f"report: {exc}")
            return None

    @staticmethod
    def _build_report_markdown(
        stats: DreamStateStats,
        assessments: List[MemoryAssessment],
        timestamp: str,
        *,
        allow_published_deprecation: bool = False,
    ) -> str:
        """Build a Markdown report of the Dream State run."""
        parts: List[str] = [
            f"# Dream State Report — {timestamp}",
            "",
            f"**Events processed:** {stats.events_processed}  ",
            f"**Memories assessed:** {stats.revisions_assessed}  ",
            f"**Duration:** {stats.duration_ms}ms",
            "",
        ]

        if allow_published_deprecation:
            parts.extend([
                "**WARNING:** Published protection was relaxed for this run "
                "(`allow_published_deprecation=true`).  ",
                "",
            ])

        parts.extend([
            "---",
            "",
            "## Actions Taken",
            "",
        ])

        # Deprecated
        deprecated = [a for a in assessments if a.should_deprecate]
        parts.append(f"### Deprecated ({stats.deprecated})")
        parts.append("")
        if deprecated:
            for a in deprecated:
                parts.append(
                    f"- `{a.revision_kref}` — {a.deprecation_reason or 'no reason given'}"
                )
        else:
            parts.append("_None_")
        parts.append("")

        # Metadata Updated
        updated = [a for a in assessments if a.metadata_updates]
        parts.append(f"### Metadata Updated ({stats.metadata_updated})")
        parts.append("")
        if updated:
            for a in updated:
                changes = ", ".join(
                    f"{k}={v}" for k, v in a.metadata_updates.items()
                )
                parts.append(f"- `{a.revision_kref}` — {changes}")
        else:
            parts.append("_None_")
        parts.append("")

        # Tags Added
        tagged = [a for a in assessments if a.suggested_tags]
        parts.append(f"### Tags Added ({stats.tags_added})")
        parts.append("")
        if tagged:
            for a in tagged:
                parts.append(
                    f"- `{a.revision_kref}` — {', '.join(a.suggested_tags)}"
                )
        else:
            parts.append("_None_")
        parts.append("")

        # Relationships Created
        related = [a for a in assessments if a.related_memories]
        parts.append(f"### Relationships Created ({stats.edges_created})")
        parts.append("")
        if related:
            for a in related:
                for target, etype in a.related_memories:
                    parts.append(
                        f"- `{a.revision_kref}` → `{target}` ({etype})"
                    )
        else:
            parts.append("_None_")
        parts.append("")

        # Errors
        if stats.errors:
            parts.append(f"### Errors ({len(stats.errors)})")
            parts.append("")
            for err in stats.errors:
                parts.append(f"- {err}")
            parts.append("")

        # Cursor
        parts.extend([
            "---",
            "",
            "## Cursor",
            "",
            f"`{stats.last_cursor or 'N/A'}`",
            "",
        ])

        return "\n".join(parts)

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------

    @staticmethod
    def _build_result(
        stats: DreamStateStats,
        *,
        report_kref: Optional[str],
    ) -> Dict[str, Any]:
        result: Dict[str, Any] = {"success": True}
        result.update(DreamState._stats_dict(stats))
        if report_kref:
            result["report_kref"] = report_kref
        return result

    @staticmethod
    def _stats_dict(stats: DreamStateStats) -> Dict[str, Any]:
        return {
            "events_processed": stats.events_processed,
            "revisions_assessed": stats.revisions_assessed,
            "deprecated": stats.deprecated,
            "metadata_updated": stats.metadata_updated,
            "tags_added": stats.tags_added,
            "edges_created": stats.edges_created,
            "cursor": stats.last_cursor,
            "duration_ms": stats.duration_ms,
            "errors": stats.errors,
        }
